from django import forms
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.core.paginator import Paginator
from django.http import HttpResponse
from django.template import TemplateDoesNotExist, TemplateSyntaxError
from django.test import RequestFactory, TestCase, override_settings

from djem.utils.tests import TemplateRendererMixin

from .models import AuditableTest


@override_settings(AUTHENTICATION_BACKENDS=[
    'django.contrib.auth.backends.ModelBackend',
    'djem.auth.ObjectPermissionsBackend'
])
class PermTagTestCase(TemplateRendererMixin, TestCase):
    
    def setUp(self):
        
        user = get_user_model().objects.create_user('test', 'fakepassword')
        
        # Grant the user all permissions for AuditableTest to ensure the tests
        # focus on object-level permissions
        permissions = Permission.objects.filter(
            content_type__app_label='djemtest',
            content_type__model='auditabletest'
        )
        user.user_permissions.set(permissions)
        
        self.user = user


class IfPermTestCase(PermTagTestCase):
    
    def test_ifperm__no_args(self):
        """
        Test the ifperm template tag raises TemplateSyntaxError when not
        provided any arguments.
        """
        
        template_string = (
            '{% load djem %}'
            '{% ifperm %}'
            '   IF'
            '{% endifperm %}'
        )
        
        with self.assertRaises(TemplateSyntaxError):
            self.render_template(template_string, {})
    
    def test_ifperm__too_few_args(self):
        """
        Test the ifperm template tag raises TemplateSyntaxError when provided
        too few arguments.
        """
        
        template_string = (
            '{% load djem %}'
            '{% ifperm "blah" %}'
            '   IF'
            '{% endifperm %}'
        )
        
        with self.assertRaises(TemplateSyntaxError):
            self.render_template(template_string, {})
    
    def test_ifperm__too_many_args(self):
        """
        Test the ifperm template tag raises TemplateSyntaxError when provided
        too many arguments.
        """
        
        template_string = (
            '{% load djem %}'
            '{% ifperm "blah" "blah" "blah" "blah" %}'
            '   IF'
            '{% endifperm %}'
        )
        
        with self.assertRaises(TemplateSyntaxError):
            self.render_template(template_string, {})
    
    def test_ifperm__invalid_permission(self):
        """
        Test the ifperm template tag does not render the "if" block of the tag,
        with no "else" block defined, when an invalid permission is provided.
        """
        
        obj = AuditableTest()
        obj.save(self.user)
        
        template_string = (
            '{% load djem %}'
            '{% ifperm user "djemtest.not_a_real_permission" obj %}'
            '   IF'
            '{% endifperm %}'
        )
        
        output = self.render_template(template_string, {
            'obj': obj
        })
        
        self.assertEqual(output, '')
    
    def test_ifperm__invalid_permission__else(self):
        """
        Test the ifperm template tag renders the "else" block of the tag, when
        defined, when an invalid permission is provided.
        """
        
        obj = AuditableTest()
        obj.save(self.user)
        
        template_string = (
            '{% load djem %}'
            '{% ifperm user "djemtest.not_a_real_permission" obj %}'
            '   IF'
            '{% else %}'
            '   ELSE'
            '{% endifperm %}'
        )
        
        output = self.render_template(template_string, {
            'obj': obj
        })
        
        self.assertEqual(output, 'ELSE')
    
    def test_ifperm__with_permission(self):
        """
        Test the ifperm template tag renders the "if" block of the tag, with no
        "else" block defined, when a user has a permission on an object.
        """
        
        obj = AuditableTest()
        obj.save(self.user)
        
        self.assertTrue(self.user.has_perm('djemtest.change_auditabletest', obj))
        
        template_string = (
            '{% load djem %}'
            '{% ifperm user "djemtest.change_auditabletest" obj %}'
            '   IF'
            '{% endifperm %}'
        )
        
        output = self.render_template(template_string, {
            'obj': obj
        })
        
        self.assertEqual(output, 'IF')
    
    def test_ifperm__with_permission__else(self):
        """
        Test the ifperm template tag renders the "if" block of the tag, with an
        "else" block defined, when a user has a permission on an object.
        """
        
        obj = AuditableTest()
        obj.save(self.user)
        
        self.assertTrue(self.user.has_perm('djemtest.change_auditabletest', obj))
        
        template_string = (
            '{% load djem %}'
            '{% ifperm user "djemtest.change_auditabletest" obj %}'
            '   IF'
            '{% else %}'
            '   ELSE'
            '{% endifperm %}'
        )
        
        output = self.render_template(template_string, {
            'obj': obj
        })
        
        self.assertEqual(output, 'IF')
    
    def test_ifperm__without_permission(self):
        """
        Test the ifperm template tag does not renders the "if" block of the tag,
        with no "else" block defined, when a user does not have a permission on
        an object.
        """
        
        other_user = get_user_model().objects.create_user('other', 'fakepassword')
        obj = AuditableTest()
        obj.save(other_user)
        
        self.assertFalse(self.user.has_perm('djemtest.change_auditabletest', obj))
        
        template_string = (
            '{% load djem %}'
            '{% ifperm user "djemtest.change_auditabletest" obj %}'
            '   IF'
            '{% endifperm %}'
        )
        
        output = self.render_template(template_string, {
            'obj': obj
        })
        
        self.assertEqual(output, '')
    
    def test_ifperm__without_permission__else(self):
        """
        Test the ifperm template tag renders the "else" block of the tag, when
        defined, when a user does not have a permission on an object.
        """
        
        other_user = get_user_model().objects.create_user('other', 'fakepassword')
        obj = AuditableTest()
        obj.save(other_user)
        
        self.assertFalse(self.user.has_perm('djemtest.change_auditabletest', obj))
        
        template_string = (
            '{% load djem %}'
            '{% ifperm user "djemtest.change_auditabletest" obj %}'
            '   IF'
            '{% else %}'
            '   ELSE'
            '{% endifperm %}'
        )
        
        output = self.render_template(template_string, {
            'obj': obj
        })
        
        self.assertEqual(output, 'ELSE')
    
    def test_ifperm__complex_user_arg(self):
        """
        Test the ifperm template tag handles the user argument being a complex
        template variable.
        """
        
        obj = AuditableTest()
        obj.save(self.user)
        
        self.assertTrue(self.user.has_perm('djemtest.change_auditabletest', obj))
        
        template_string = (
            '{% load djem %}'
            '{% ifperm some.complex.variable "djemtest.change_auditabletest" obj %}'
            '   IF'
            '{% else %}'
            '   ELSE'
            '{% endifperm %}'
        )
        
        output = self.render_template(template_string, {
            'some': {
                'complex': {
                    'variable': self.user
                }
            },
            'obj': obj
        })
        
        self.assertEqual(output, 'IF')
    
    def test_ifperm__complex_permission_arg(self):
        """
        Test the ifperm template tag handles the permission argument being a
        complex template variable.
        """
        
        obj = AuditableTest()
        obj.save(self.user)
        
        self.assertTrue(self.user.has_perm('djemtest.change_auditabletest', obj))
        
        template_string = (
            '{% load djem %}'
            '{% ifperm user some.complex.variable obj %}'
            '   IF'
            '{% else %}'
            '   ELSE'
            '{% endifperm %}'
        )
        
        output = self.render_template(template_string, {
            'obj': obj,
            'some': {
                'complex': {
                    'variable': 'djemtest.change_auditabletest'
                }
            }
        })
        
        self.assertEqual(output, 'IF')
    
    def test_ifperm__complex_object_arg(self):
        """
        Test the ifperm template tag handles the object argument being a
        complex template variable.
        """
        
        obj = AuditableTest()
        obj.save(self.user)
        
        self.assertTrue(self.user.has_perm('djemtest.change_auditabletest', obj))
        
        template_string = (
            '{% load djem %}'
            '{% ifperm user "djemtest.change_auditabletest" some.complex.variable %}'
            '   IF'
            '{% else %}'
            '   ELSE'
            '{% endifperm %}'
        )
        
        output = self.render_template(template_string, {
            'some': {
                'complex': {
                    'variable': obj
                }
            }
        })
        
        self.assertEqual(output, 'IF')
    
    def test_ifperm__view__anonymous(self):
        """
        Test the ifperm template tag when used in a template rendered as part of
        a request/response cycle (i.e. standard real world usage), when there
        is no authenticated user.
        """
        
        def view(r, o):
            
            template_string = (
                '{% load djem %}'
                '{% ifperm user "djemtest.change_auditabletest" obj %}'
                '   IF'
                '{% else %}'
                '   ELSE'
                '{% endifperm %}'
            )
            
            output = self.render_template(template_string, {
                'obj': o
            }, r)
            
            return HttpResponse(output)
        
        obj = AuditableTest()
        obj.save(self.user)
        
        request = RequestFactory().get('/test/')
        response = view(request, obj)
        
        self.assertContains(response, 'ELSE', status_code=200)
    
    def test_ifperm__view__login(self):
        """
        Test the ifperm template tag when used in a template rendered as part of
        a request/response cycle (i.e. standard real world usage), when there
        is an authenticated user.
        """
        
        def view(r, o):
            
            template_string = (
                '{% load djem %}'
                '{% ifperm user "djemtest.change_auditabletest" obj %}'
                '   IF'
                '{% else %}'
                '   ELSE'
                '{% endifperm %}'
            )
            
            output = self.render_template(template_string, {
                'obj': o
            }, r)
            
            return HttpResponse(output)
        
        obj = AuditableTest()
        obj.save(self.user)
        
        request = RequestFactory().get('/test/')
        request.user = self.user  # simulate login
        response = view(request, obj)
        
        self.assertContains(response, 'IF', status_code=200)


class IfNotPermTestCase(PermTagTestCase):
    
    def test_ifnotperm__no_args(self):
        """
        Test the ifnotperm template tag raises TemplateSyntaxError when not
        provided any arguments.
        """
        
        template_string = (
            '{% load djem %}'
            '{% ifnotperm %}'
            '   IF'
            '{% endifnotperm %}'
        )
        
        with self.assertRaises(TemplateSyntaxError):
            self.render_template(template_string, {})
    
    def test_ifnotperm__too_few_args(self):
        """
        Test the ifnotperm template tag raises TemplateSyntaxError when provided
        too few arguments.
        """
        
        template_string = (
            '{% load djem %}'
            '{% ifnotperm "blah" %}'
            '   IF'
            '{% endifnotperm %}'
        )
        
        with self.assertRaises(TemplateSyntaxError):
            self.render_template(template_string, {})
    
    def test_ifnotperm__too_many_args(self):
        """
        Test the ifnotperm template tag raises TemplateSyntaxError when provided
        too many arguments.
        """
        
        template_string = (
            '{% load djem %}'
            '{% ifnotperm "blah" "blah" "blah" "blah" %}'
            '   IF'
            '{% endifnotperm %}'
        )
        
        with self.assertRaises(TemplateSyntaxError):
            self.render_template(template_string, {})
    
    def test_ifnotperm__invalid_permission(self):
        """
        Test the ifnotperm template tag renders the "if" block of the tag, with
        no "else" block defined, when an invalid permission is provided.
        """
        
        obj = AuditableTest()
        obj.save(self.user)
        
        template_string = (
            '{% load djem %}'
            '{% ifnotperm user "djemtest.not_a_real_permission" obj %}'
            '   IF'
            '{% endifnotperm %}'
        )
        
        output = self.render_template(template_string, {
            'obj': obj
        })
        
        self.assertEqual(output, 'IF')
    
    def test_ifnotperm__invalid_permission__else(self):
        """
        Test the ifnotperm template tag renders the "if" block of the tag, when
        defined, when an invalid permission is provided.
        """
        
        obj = AuditableTest()
        obj.save(self.user)
        
        template_string = (
            '{% load djem %}'
            '{% ifnotperm user "djemtest.not_a_real_permission" obj %}'
            '   IF'
            '{% else %}'
            '   ELSE'
            '{% endifnotperm %}'
        )
        
        output = self.render_template(template_string, {
            'obj': obj
        })
        
        self.assertEqual(output, 'IF')
    
    def test_ifnotperm__with_permission(self):
        """
        Test the ifnotperm template tag does not render the "if" block of the
        tag, with no "else" block defined, when a user has a permission on an
        object.
        """
        
        obj = AuditableTest()
        obj.save(self.user)
        
        self.assertTrue(self.user.has_perm('djemtest.change_auditabletest', obj))
        
        template_string = (
            '{% load djem %}'
            '{% ifnotperm user "djemtest.change_auditabletest" obj %}'
            '   IF'
            '{% endifnotperm %}'
        )
        
        output = self.render_template(template_string, {
            'obj': obj
        })
        
        self.assertEqual(output, '')
    
    def test_ifnotperm__with_permission__else(self):
        """
        Test the ifnotperm template tag renders the "else" block of the tag,
        when defined, when a user has a permission on an object.
        """
        
        obj = AuditableTest()
        obj.save(self.user)
        
        self.assertTrue(self.user.has_perm('djemtest.change_auditabletest', obj))
        
        template_string = (
            '{% load djem %}'
            '{% ifnotperm user "djemtest.change_auditabletest" obj %}'
            '   IF'
            '{% else %}'
            '   ELSE'
            '{% endifnotperm %}'
        )
        
        output = self.render_template(template_string, {
            'obj': obj
        })
        
        self.assertEqual(output, 'ELSE')
    
    def test_ifnotperm__without_permission(self):
        """
        Test the ifnotperm template tag renders the "if" block of the tag, with
        with no "else" block defined, when a user does not have a permission on
        an object.
        """
        
        other_user = get_user_model().objects.create_user('other', 'fakepassword')
        obj = AuditableTest()
        obj.save(other_user)
        
        self.assertFalse(self.user.has_perm('djemtest.change_auditabletest', obj))
        
        template_string = (
            '{% load djem %}'
            '{% ifnotperm user "djemtest.change_auditabletest" obj %}'
            '   IF'
            '{% endifnotperm %}'
        )
        
        output = self.render_template(template_string, {
            'obj': obj
        })
        
        self.assertEqual(output, 'IF')
    
    def test_ifnotperm__without_permission__else(self):
        """
        Test the ifnotperm template tag renders the "if" block of the tag, with
        with an "else" block defined, when a user does not have a permission on
        an object.
        """
        
        other_user = get_user_model().objects.create_user('other', 'fakepassword')
        obj = AuditableTest()
        obj.save(other_user)
        
        self.assertFalse(self.user.has_perm('djemtest.change_auditabletest', obj))
        
        template_string = (
            '{% load djem %}'
            '{% ifnotperm user "djemtest.change_auditabletest" obj %}'
            '   IF'
            '{% else %}'
            '   ELSE'
            '{% endifnotperm %}'
        )
        
        output = self.render_template(template_string, {
            'obj': obj
        })
        
        self.assertEqual(output, 'IF')
    
    def test_ifperm__complex_user_arg(self):
        """
        Test the ifperm template tag handles the user argument being a complex
        template variable.
        """
        
        obj = AuditableTest()
        obj.save(self.user)
        
        self.assertTrue(self.user.has_perm('djemtest.change_auditabletest', obj))
        
        template_string = (
            '{% load djem %}'
            '{% ifnotperm some.complex.variable "djemtest.change_auditabletest" obj %}'
            '   IF'
            '{% else %}'
            '   ELSE'
            '{% endifnotperm %}'
        )
        
        output = self.render_template(template_string, {
            'some': {
                'complex': {
                    'variable': self.user
                }
            },
            'obj': obj
        })
        
        self.assertEqual(output, 'ELSE')
    
    def test_ifnotperm__complex_permission_arg(self):
        """
        Test the ifnotperm template tag handles the permission argument being a
        complex template variable.
        """
        
        obj = AuditableTest()
        obj.save(self.user)
        
        self.assertTrue(self.user.has_perm('djemtest.change_auditabletest', obj))
        
        template_string = (
            '{% load djem %}'
            '{% ifnotperm user some.complex.variable obj %}'
            '   IF'
            '{% else %}'
            '   ELSE'
            '{% endifnotperm %}'
        )
        
        output = self.render_template(template_string, {
            'obj': obj,
            'some': {
                'complex': {
                    'variable': 'djemtest.change_auditabletest'
                }
            }
        })
        
        self.assertEqual(output, 'ELSE')
    
    def test_ifnotperm__complex_object_arg(self):
        """
        Test the ifnotperm template tag handles the object argument being a
        complex template variable.
        """
        
        obj = AuditableTest()
        obj.save(self.user)
        
        self.assertTrue(self.user.has_perm('djemtest.change_auditabletest', obj))
        
        template_string = (
            '{% load djem %}'
            '{% ifnotperm user "djemtest.change_auditabletest" some.complex.variable %}'
            '   IF'
            '{% else %}'
            '   ELSE'
            '{% endifnotperm %}'
        )
        
        output = self.render_template(template_string, {
            'some': {
                'complex': {
                    'variable': obj
                }
            }
        })
        
        self.assertEqual(output, 'ELSE')
    
    def test_ifnotperm__view__anonymous(self):
        """
        Test the ifnotperm template tag when used in a template rendered as part
        of a request/response cycle (i.e. standard real world usage), when there
        is no authenticated user.
        """
        
        def view(r, o):
            
            template_string = (
                '{% load djem %}'
                '{% ifnotperm user "djemtest.change_auditabletest" obj %}'
                '   IF'
                '{% else %}'
                '   ELSE'
                '{% endifnotperm %}'
            )
            
            output = self.render_template(template_string, {
                'obj': o
            }, r)
            
            return HttpResponse(output)
        
        obj = AuditableTest()
        obj.save(self.user)
        
        request = RequestFactory().get('/test/')
        response = view(request, obj)
        
        self.assertContains(response, 'IF', status_code=200)
    
    def test_ifnotperm__view__login(self):
        """
        Test the ifnotperm template tag when used in a template rendered as part
        of a request/response cycle (i.e. standard real world usage), when there
        is an authenticated user.
        """
        
        def view(r, o):
            
            template_string = (
                '{% load djem %}'
                '{% ifnotperm user "djemtest.change_auditabletest" obj %}'
                '   IF'
                '{% else %}'
                '   ELSE'
                '{% endifnotperm %}'
            )
            
            output = self.render_template(template_string, {
                'obj': o
            }, r)
            
            return HttpResponse(output)
        
        obj = AuditableTest()
        obj.save(self.user)
        
        request = RequestFactory().get('/test/')
        request.user = self.user  # simulate login
        response = view(request, obj)
        
        self.assertContains(response, 'ELSE', status_code=200)


class CsrfifyAjaxTestCase(TemplateRendererMixin, TestCase):
    
    def test_valid__explicit(self):
        """
        Test the csrfify_ajax template tag when used in a request/response
        cycle with a present CSRF token, and with an explicitly provided valid
        target library.
        """
        
        def view(r):
            
            template_string = "{% load djem %}\n{% csrfify_ajax 'jquery' %}"
            
            output = self.render_template(template_string, {}, r)
            
            return HttpResponse(output)
        
        request = RequestFactory().get('/test/')
        response = view(request)
        
        self.assertEqual(response.status_code, 200)
        self.assertRegex(response.content, b"('X-CSRFToken', '[a-zA-Z0-9]{64}')")
    
    def test_valid__implicit(self):
        """
        Test the csrfify_ajax template tag when used in a request/response
        cycle with a present CSRF token, and with no arguments (therefore using
        the default target library).
        """
        
        def view(r):
            
            template_string = "{% load djem %}\n{% csrfify_ajax %}"
            
            output = self.render_template(template_string, {}, r)
            
            return HttpResponse(output)
        
        request = RequestFactory().get('/test/')
        response = view(request)
        
        self.assertEqual(response.status_code, 200)
        self.assertRegex(response.content, b"('X-CSRFToken', '[a-zA-Z0-9]{64}')")
    
    def test_invalid(self):
        """
        Test the csrfify_ajax template tag when used in a request/response
        cycle with a present CSRF token, and with an explicitly provided but
        invalid target library.
        """
        
        def view(r):
            
            template_string = "{% load djem %}\n{% csrfify_ajax 'invalid' %}"
            
            self.render_template(template_string, {}, r)
        
        request = RequestFactory().get('/test/')
        
        with self.assertRaisesMessage(TemplateDoesNotExist, 'invalid.html'):
            view(request)


class PaginateTestCase(TemplateRendererMixin, TestCase):
    
    @classmethod
    def setUpTestData(cls):
        
        user = get_user_model().objects.create_user('test')
        
        for _ in range(23):
            AuditableTest().save(user)
        
        # All tests will use the same Paginator with 5 results per page
        cls.paginator = Paginator(AuditableTest.objects.order_by('pk'), 5)
        
        # All tests will render the same basic template - just the
        # {% paginate %} tag on the given Page. It is the Page object that will
        # differ between tests.
        cls.template_string = (
            '{% load djem %}'
            '{% paginate page %}'
        )
        
        cls.user = user  # for TemplateRendererMixin
    
    def test_no_argument(self):
        """
        Test the paginate template tag when not provided with an argument. It
        should raise a TemplateSyntaxError - a single argument is required.
        """
        
        template_string = (
            '{% load djem %}'
            '{% paginate %}'
        )
        
        with self.assertRaises(TemplateSyntaxError):
            self.render_template(template_string, {})
    
    def test_multiple_arguments(self):
        """
        Test the paginate template tag when provided with multiple arguments. It
        should raise a TemplateSyntaxError - a single argument is required.
        """
        
        template_string = (
            '{% load djem %}'
            '{% paginate "multiple" "arguments" %}'
        )
        
        with self.assertRaises(TemplateSyntaxError):
            self.render_template(template_string, {})
    
    def test_first_page(self):
        """
        Test the paginate template tag when used on the first Page of a
        Paginator. It should render without previous/first links.
        """
        
        page = self.paginator.page(1)
        
        output = self.render_template(self.template_string, {'page': page})
        
        # The pagination block for the first page should:
        #  - show the page number and the number of pages
        #  - contain links to "Next" and "Last"
        #  - not contain links to "Previous" or "First"
        self.assertIn('Page 1 of 5', output)
        
        self.assertIn('"?page=2">Next', output)
        self.assertIn('"?page=5">Last', output)
        
        self.assertNotIn('Previous', output)
        self.assertNotIn('First', output)
    
    def test_last_page(self):
        """
        Test the paginate template tag when used on the last Page of a
        Paginator. It should render without next/last links.
        """
        
        page = self.paginator.page(5)
        
        output = self.render_template(self.template_string, {'page': page})
        
        # The pagination block for the first page should:
        #  - show the page number and the number of pages
        #  - contain links to "Previous" and "First"
        #  - not contain links to "Next" or "Last"
        self.assertIn('Page 5 of 5', output)
        
        self.assertIn('"?page=4">Previous', output)
        self.assertIn('"?page=1">First', output)
        
        self.assertNotIn('Next', output)
        self.assertNotIn('Last', output)
    
    def test_middle_page(self):
        """
        Test the paginate template tag when used on a middle Page of a
        Paginator. It should render with all navigation links.
        """
        
        page = self.paginator.page(3)
        
        output = self.render_template(self.template_string, {'page': page})
        
        # The pagination block for the first page should:
        #  - show the page number and the number of pages
        #  - contain links to "Previous" and "First"
        #  - contain links to "Next" and "Last"
        self.assertIn('Page 3 of 5', output)
        
        self.assertIn('"?page=2">Previous', output)
        self.assertIn('"?page=1">First', output)
        
        self.assertIn('"?page=4">Next', output)
        self.assertIn('"?page=5">Last', output)
    
    def test_invalid_page(self):
        """
        Test the paginate template tag when given something other than a Page
        object. Due to not having the attributes expected on a Page object, it
        should render without navigation links or valid page numbers.
        """
        
        output = self.render_template(self.template_string, {'page': None}, flatten=False)
        
        self.assertIn('Page  of ', output)
        
        self.assertNotIn('Previous', output)
        self.assertNotIn('First', output)
        
        self.assertNotIn('Next', output)
        self.assertNotIn('Last', output)


class FormFieldTestCase(TemplateRendererMixin, TestCase):
    
    @classmethod
    def setUpTestData(cls):
        
        class TestForm(forms.Form):
            
            required_css_class = 'form-field--required'
            
            field = forms.CharField(required=False, label='Test Field')
            required_field = forms.CharField()
            help_field = forms.CharField(help_text='Some helpful text.')
        
        cls.form = TestForm
    
    def test_no_argument(self):
        """
        Test the form_field template tag when not provided with an argument. It
        should raise a TemplateSyntaxError - at least one argument is required.
        """
        
        template_string = (
            '{% load djem %}'
            '{% form_field %}'
        )
        
        with self.assertRaises(TemplateSyntaxError):
            self.render_template(template_string, {})
    
    def test_default_wrapper(self):
        """
        Test the form_field template tag when no custom ``DJEM_FORM_FIELD_TAG``
        setting has been specified. It should use a <div> as the wrapping
        element.
        """
        
        template_string = (
            '{% load djem %}'
            '{% form_field form.field %}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form()
        })
        
        self.assertEqual(output[:25], '<div class="form-field" >')
        self.assertEqual(output[-6:], '</div>')
    
    @override_settings(DJEM_FORM_FIELD_TAG='li')
    def test_custom_wrapper(self):
        """
        Test the form_field template tag when a custom ``DJEM_FORM_FIELD_TAG``
        setting has been specified. It should use the specified tag as the
        wrapping element.
        """
        
        template_string = (
            '{% load djem %}'
            '{% form_field form.field %}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form()
        })
        
        self.assertEqual(output[:24], '<li class="form-field" >')
        self.assertEqual(output[-5:], '</li>')
    
    def test_content(self):
        """
        Test the form_field template tag in basic usage. It should output all
        expected HTML.
        """
        
        template_string = (
            '{% load djem %}'
            '{% form_field form.field %}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form()
        })
        
        expected_output = (
            '<div class="form-field" >'
            '<label for="id_field">Test Field:</label>'
            '<input type="text" name="field" id="id_field">'
            '</div>'
        )
        
        self.assertEqual(output, expected_output)
    
    def test_errors(self):
        """
        Test the form_field template tag when used with a field that contains
        errors after a form submission. It should add an element containing the
        list of errors.
        """
        
        template_string = (
            '{% load djem %}'
            '{% form_field form.required_field %}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form()
        })
        
        self.assertNotIn('<ul class="errors">', output)
        
        template_string = (
            '{% load djem %}'
            '{% form_field form.required_field %}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form({})  # bind form
        })
        
        self.assertIn('<ul class="errorlist"', output)
        self.assertIn('<li>This field is required.</li>', output)
    
    def test_help_text(self):
        """
        Test the form_field template tag when used with a field that defines
        help text. It should add an element containing the help text.
        """
        
        form = self.form()
        template_string = (
            '{% load djem %}'
            '{% form_field form.field %}'
        )
        
        output = self.render_template(template_string, {
            'form': form
        })
        
        self.assertNotIn('<div class="form-field__help">', output)
        
        template_string = (
            '{% load djem %}'
            '{% form_field form.help_field %}'
        )
        
        output = self.render_template(template_string, {
            'form': form
        })
        
        self.assertIn('<div class="form-field__help">Some helpful text.</div>', output)
    
    def test_css_classes(self):
        """
        Test the form_field template tag when used with a field that defines
        css classes. It should add those classes to the wrapping element.
        """
        
        template_string = (
            '{% load djem %}'
            '{% form_field form.required_field %}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form()
        })
        
        self.assertIn('class="form-field form-field--required"', output)
    
    def test_extra_classes__string(self):
        """
        Test the form_field template tag when the ``extra_classes`` argument is
        given as a string. It should include those classes on the wrapping
        element.
        """
        
        template_string = (
            '{% load djem %}'
            '{% form_field form.field "testclass1 testclass2" %}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form()
        })
        
        self.assertIn('class="form-field testclass1 testclass2"', output)
    
    def test_extra_classes__variable(self):
        """
        Test the form_field template tag when the ``extra_classes`` argument is
        given as a template context variable. It should include those classes
        on the wrapping element.
        """
        
        template_string = (
            '{% load djem %}'
            '{% form_field form.required_field classes %}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form(),
            'classes': 'testclass1 testclass2'
        })
        
        self.assertIn('class="form-field form-field--required testclass1 testclass2"', output)
    
    def test_extra_attributes__strings(self):
        """
        Test the form_field template tag when keyword arguments are given with
        values as strings. It should include those arguments as HTML attributes
        on the wrapping element, replacing underscores with dashes as necessary.
        """
        
        template_string = (
            '{% load djem %}'
            '{% form_field form.field data_id="1234" %}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form()
        })
        
        self.assertIn('<div class="form-field" data-id="1234" >', output)
    
    def test_extra_attributes__variables(self):
        """
        Test the form_field template tag when keyword arguments are given with
        values as template context variables. It should include those arguments
        as HTML attributes on the wrapping element, replacing underscores with
        dashes as necessary.
        """
        
        template_string = (
            '{% load djem %}'
            '{% form_field form.field data_id=id %}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form(),
            'id': '1234'
        })
        
        self.assertIn('<div class="form-field" data-id="1234" >', output)


class CheckboxTestCase(TemplateRendererMixin, TestCase):
    
    @classmethod
    def setUpTestData(cls):
        
        class TestForm(forms.Form):
            
            required_css_class = 'form-field--required'
            
            field = forms.BooleanField(required=False, label='Test Field')
            required_field = forms.BooleanField()
            help_field = forms.BooleanField(help_text='Some helpful text.')
        
        cls.form = TestForm
    
    def test_no_argument(self):
        """
        Test the checkbox template tag when not provided with an argument. It
        should raise a TemplateSyntaxError - at least one argument is required.
        """
        
        template_string = (
            '{% load djem %}'
            '{% checkbox %}{% endcheckbox%}'
        )
        
        with self.assertRaisesMessage(TemplateSyntaxError, "'checkbox' takes at least one argument"):
            self.render_template(template_string, {})
    
    def test_no_end_tag(self):
        """
        Test the checkbox template tag when not provided with an argument. It
        should raise a TemplateSyntaxError - at least one argument is required.
        """
        
        template_string = (
            '{% load djem %}'
            '{% checkbox form.field %}'
        )
        
        with self.assertRaisesMessage(TemplateSyntaxError, 'Unclosed tag'):
            self.render_template(template_string, {
                'form': self.form()
            })
    
    def test_default_wrapper(self):
        """
        Test the checkbox template tag when no custom ``DJEM_FORM_FIELD_TAG``
        setting has been specified. It should use a <div> as the wrapping
        element.
        """
        
        template_string = (
            '{% load djem %}'
            '{% checkbox form.field %}Click this{% endcheckbox%}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form()
        })
        
        self.assertEqual(output[:25], '<div class="form-field" >')
        self.assertEqual(output[-6:], '</div>')
    
    @override_settings(DJEM_FORM_FIELD_TAG='li')
    def test_custom_wrapper(self):
        """
        Test the checkbox template tag when a custom ``DJEM_FORM_FIELD_TAG``
        setting has been specified. It should use the specified tag as the
        wrapping element.
        """
        
        template_string = (
            '{% load djem %}'
            '{% checkbox form.field %}Click this{% endcheckbox%}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form()
        })
        
        self.assertEqual(output[:24], '<li class="form-field" >')
        self.assertEqual(output[-5:], '</li>')
    
    def test_content(self):
        """
        Test the checkbox template tag in basic usage. It should output all
        expected HTML.
        """
        
        template_string = (
            '{% load djem %}'
            '{% checkbox form.field %}Click this{% endcheckbox%}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form()
        })
        
        expected_output = (
            '<div class="form-field" >'
            '<input type="checkbox" name="field" id="id_field">'
            '<label class="check-label" for="id_field"> Click this </label>'
            '</div>'
        )
        
        self.assertEqual(output, expected_output)
    
    def test_content__empty(self):
        """
        Test the checkbox template tag when no content in entered between the
        start and end tags. It should use the field's own label.
        """
        
        template_string = (
            '{% load djem %}'
            '{% checkbox form.field %}{% endcheckbox%}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form()
        })
        
        expected_output = (
            '<div class="form-field" >'
            '<input type="checkbox" name="field" id="id_field">'
            '<label class="check-label" for="id_field"> Test Field </label>'
            '</div>'
        )
        
        self.assertEqual(output, expected_output)
    
    def test_errors(self):
        """
        Test the checkbox template tag when used with a field that contains
        errors after a form submission. It should add an element containing the
        list of errors.
        """
        
        template_string = (
            '{% load djem %}'
            '{% checkbox form.required_field %}Click this{% endcheckbox%}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form()
        })
        
        self.assertNotIn('<ul class="errors">', output)
        
        template_string = (
            '{% load djem %}'
            '{% checkbox form.required_field %}Click this{% endcheckbox%}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form({})  # bind form
        })
        
        self.assertIn('<ul class="errorlist"', output)
        self.assertIn('<li>This field is required.</li>', output)
    
    def test_help_text(self):
        """
        Test the checkbox template tag when used with a field that defines
        help text. It should add an element containing the help text.
        """
        
        form = self.form()
        template_string = (
            '{% load djem %}'
            '{% checkbox form.field %}Click this{% endcheckbox%}'
        )
        
        output = self.render_template(template_string, {
            'form': form
        })
        
        self.assertNotIn('<div class="form-field__help">', output)
        
        template_string = (
            '{% load djem %}'
            '{% checkbox form.help_field %}Click this{% endcheckbox%}'
        )
        
        output = self.render_template(template_string, {
            'form': form
        })
        
        self.assertIn('<div class="form-field__help">Some helpful text.</div>', output)
    
    def test_css_classes(self):
        """
        Test the checkbox template tag when used with a field that defines
        css classes. It should add those classes to the wrapping element.
        """
        
        template_string = (
            '{% load djem %}'
            '{% checkbox form.required_field %}Click this{% endcheckbox%}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form()
        })
        
        self.assertIn('class="form-field form-field--required"', output)
    
    def test_extra_classes__string(self):
        """
        Test the checkbox template tag when the ``extra_classes`` argument is
        given as a string. It should include those classes on the wrapping
        element.
        """
        
        template_string = (
            '{% load djem %}'
            '{% checkbox form.field "testclass1 testclass2" %}Click this{% endcheckbox%}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form()
        })
        
        self.assertIn('class="form-field testclass1 testclass2"', output)
    
    def test_extra_classes__variable(self):
        """
        Test the checkbox template tag when the ``extra_classes`` argument is
        given as a template context variable. It should include those classes
        on the wrapping element.
        """
        
        template_string = (
            '{% load djem %}'
            '{% checkbox form.required_field classes %}Click this{% endcheckbox%}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form(),
            'classes': 'testclass1 testclass2'
        })
        
        self.assertIn('class="form-field form-field--required testclass1 testclass2"', output)
    
    def test_extra_attributes__strings(self):
        """
        Test the checkbox template tag when keyword arguments are given with
        values as strings. It should include those arguments as HTML attributes
        on the wrapping element, replacing underscores with dashes as necessary.
        """
        
        template_string = (
            '{% load djem %}'
            '{% checkbox form.field data_id="1234" %}Click this{% endcheckbox%}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form()
        })
        
        self.assertIn('<div class="form-field" data-id="1234" >', output)
    
    def test_extra_attributes__variables(self):
        """
        Test the checkbox template tag when keyword arguments are given with
        values as template context variables. It should include those arguments
        as HTML attributes on the wrapping element, replacing underscores with
        dashes as necessary.
        """
        
        template_string = (
            '{% load djem %}'
            '{% checkbox form.field data_id=id %}Click this{% endcheckbox%}'
        )
        
        output = self.render_template(template_string, {
            'form': self.form(),
            'id': '1234'
        })
        
        self.assertIn('<div class="form-field" data-id="1234" >', output)



from django.contrib.auth.models import AbstractUser, Group, Permission
from django.core.exceptions import PermissionDenied
from django.db import models

from djem.models import (
    Archivable,
    Auditable,
    Loggable,
    OLPMixin,
    StaticAbstract,
    TimeZoneField,
    Versionable,
)


class AuditableTest(Auditable, models.Model):
    """
    This model provides a concrete model with the Auditable mixin for testing.
    """
    
    field1 = models.BooleanField(default=True)
    field2 = models.BooleanField(default=True)
    
    def _user_can_change_auditabletest(self, user):
        
        return user.pk == self.user_created_id


class ArchivableTest(Archivable, models.Model):
    """
    This model provides a concrete model with the Archivable for testing.
    """
    
    field1 = models.BooleanField(default=True)
    field2 = models.BooleanField(default=True)


class VersionableTest(Versionable, models.Model):
    """
    This model provides a concrete model with the Versionable mixin for testing.
    """
    
    field1 = models.BooleanField(default=True)
    field2 = models.BooleanField(default=True)


class StaticTest(StaticAbstract):
    """
    This model provides a concrete version of the abstract StaticAbstract
    model for testing.
    """
    
    field1 = models.BooleanField(default=True)
    field2 = models.BooleanField(default=True)


class RelatedArchivableTest(Archivable, models.Model):
    """
    This model provides a concrete model with relationships to other test
    models, for testing protected/restricted foreign key detection in
    ``archive()`` (for archivable related records).
    """
    
    # FKs to ArchivableTest
    archivable_protected = models.ForeignKey(ArchivableTest, on_delete=models.PROTECT, null=True, related_name='+')
    archivable_restricted = models.ForeignKey(ArchivableTest, on_delete=models.RESTRICT, null=True, related_name='+')
    
    # FKs to AuditableTest
    auditable_protected = models.ForeignKey(AuditableTest, on_delete=models.PROTECT, null=True, related_name='+')


class RelatedTest(models.Model):
    """
    This model provides a concrete model with relationships to other test
    models, for testing protected/restricted foreign key detection in
    ``archive()`` (for non-archivable related records).
    """
    
    # FKs to ArchivableTest
    archivable_protected = models.ForeignKey(ArchivableTest, on_delete=models.PROTECT, null=True, related_name='+')
    archivable_restricted = models.ForeignKey(ArchivableTest, on_delete=models.RESTRICT, null=True, related_name='+')
    
    # FKs to AuditableTest
    auditable_protected = models.ForeignKey(AuditableTest, on_delete=models.PROTECT, null=True, related_name='+')


class TimeZoneTest(models.Model):
    """
    This model provides some TimeZoneFields for testing.
    """
    
    timezone = TimeZoneField()
    timezone2 = TimeZoneField(default='Australia/Sydney')
    timezone3 = TimeZoneField(null=True)
    timezone4 = TimeZoneField(max_length=32, choices=(
        ('Australia/Sydney', 'Australia/Sydney'),
        ('Australia/Melbourne', 'Australia/Melbourne'),
        ('invalid', 'Invalid'),
    ))


class LogTest(Loggable, models.Model):
    """
    This model provides a concrete model with the Loggable mixin for testing.
    """
    
    field = models.BooleanField(default=True)


class CustomUser(OLPMixin, AbstractUser):
    """
    This is a custom user model for testing OLPMixin.
    """
    
    # Override groups and user_permissions to avoid related_name clashes with
    # the same fields on the regular auth.User model
    groups = models.ManyToManyField(Group, related_name='+')
    user_permissions = models.ManyToManyField(Permission, related_name='+')


class UserLogTest(models.Model):
    """
    This is a contrived model for testing OLPMixin permission logging.
    """
    
    class Meta:
        permissions = (
            ('mlp_log', 'Permission with no object-level checks'),
            ('olp_log', 'Permission with object-level checks and additional logging'),
        )
        
        # For test consistency across versions - due to the lack of a default
        # "view" permission prior to Django 2.1
        default_permissions = ('view', 'add', 'change', 'delete')
    
    def __str__(self):
        
        return f'Log Test #{self.pk}'
    
    def _user_can_olp_log(self, user):
        
        user.log('This permission is restricted.')
        
        return False


class OLPTest(models.Model):
    """
    This is a contrived model for testing object-level permissions with the
    standard user model.
    """
    
    user = models.ForeignKey('auth.User', null=True, on_delete=models.PROTECT)
    group = models.ForeignKey(Group, null=True, on_delete=models.PROTECT)
    
    class Meta:
        permissions = (
            ('open_olptest', 'Completely open to anyone'),
            ('closed_olptest', 'Completely closed to everyone (except super users)'),
            ('user_only_olptest', 'Open to the user specified on the object'),
            ('group_only_olptest', 'Open to the group specified on the object'),
            ('combined_olptest', 'Open to any user OR group specified on the object'),
            ('deny_olptest', 'Denies permission by raising PermissionDenied for the user'),
        )
        
        # For test consistency across versions - due to the lack of a default
        # "view" permission prior to Django 2.1
        default_permissions = ('view', 'add', 'change', 'delete')
    
    def _user_can_open_olptest(self, user):
        
        # For tests using inactive and super users (neither should reach this point)
        assert user.is_active or not user.is_superuser, 'Not supposed to get here'  # noqa: S101
        
        return True
    
    def _group_can_open_olptest(self, groups):
        
        return True
    
    def _user_can_user_only_olptest(self, user):
        
        return user.pk == self.user_id
    
    def _group_can_user_only_olptest(self, groups):
        
        return False
    
    def _user_can_group_only_olptest(self, groups):
        
        return False
    
    def _group_can_group_only_olptest(self, groups):
        
        return groups.filter(pk=self.group_id).exists()
    
    def _user_can_combined_olptest(self, user):
        
        return user.pk == self.user_id
    
    def _group_can_combined_olptest(self, groups):
        
        return groups.filter(pk=self.group_id).exists()
    
    def _user_can_deny_olptest(self, user):
        
        raise PermissionDenied

 
class UniversalOLPTest(models.Model):
    """
    This is a contrived model for testing object-level permissions with the
    above custom user model incorporating OLPMixin.
    """
    
    user = models.ForeignKey(CustomUser, null=True, on_delete=models.PROTECT)
    group = models.ForeignKey(Group, null=True, on_delete=models.PROTECT)
    
    class Meta:
        permissions = (
            ('open_universalolptest', 'Completely open to anyone'),
            ('closed_universalolptest', 'Completely closed to everyone (except super users)'),
            ('user_only_universalolptest', 'Open to the user specified on the object'),
            ('group_only_universalolptest', 'Open to the group specified on the object'),
            ('combined_universalolptest', 'Open to any user OR group specified on the object'),
            ('deny_universalolptest', 'Denies permission by raising PermissionDenied for the user'),
        )
        
        # For test consistency across versions - due to the lack of a default
        # "view" permission prior to Django 2.1
        default_permissions = ('view', 'add', 'change', 'delete')
    
    def _user_can_open_universalolptest(self, user):
        
        # For tests using inactive and super users (neither should reach this point)
        assert user.is_active or not user.is_superuser, 'Not supposed to get here'  # noqa: S101
        
        return True
    
    def _group_can_open_universalolptest(self, groups):
        
        return True
    
    def _user_can_user_only_universalolptest(self, user):
        
        return user.pk == self.user_id
    
    def _group_can_user_only_universalolptest(self, groups):
        
        return False
    
    def _user_can_group_only_universalolptest(self, groups):
        
        return False
    
    def _group_can_group_only_universalolptest(self, groups):
        
        return groups.filter(pk=self.group_id).exists()
    
    def _user_can_combined_universalolptest(self, user):
        
        return user.pk == self.user_id
    
    def _group_can_combined_universalolptest(self, groups):
        
        return groups.filter(pk=self.group_id).exists()
    
    def _user_can_deny_universalolptest(self, user):
        
        raise PermissionDenied



# Date, time and timezone utils
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

from django.core.exceptions import ValidationError
from django.utils import timezone

# Pytz's "common timezones" list based on the 2024b IANA timezone database
TIMEZONE_CHOICES = [(tz, tz) for tz in [
    'Africa/Abidjan',
    'Africa/Accra',
    'Africa/Addis_Ababa',
    'Africa/Algiers',
    'Africa/Asmara',
    'Africa/Bamako',
    'Africa/Bangui',
    'Africa/Banjul',
    'Africa/Bissau',
    'Africa/Blantyre',
    'Africa/Brazzaville',
    'Africa/Bujumbura',
    'Africa/Cairo',
    'Africa/Casablanca',
    'Africa/Ceuta',
    'Africa/Conakry',
    'Africa/Dakar',
    'Africa/Dar_es_Salaam',
    'Africa/Djibouti',
    'Africa/Douala',
    'Africa/El_Aaiun',
    'Africa/Freetown',
    'Africa/Gaborone',
    'Africa/Harare',
    'Africa/Johannesburg',
    'Africa/Juba',
    'Africa/Kampala',
    'Africa/Khartoum',
    'Africa/Kigali',
    'Africa/Kinshasa',
    'Africa/Lagos',
    'Africa/Libreville',
    'Africa/Lome',
    'Africa/Luanda',
    'Africa/Lubumbashi',
    'Africa/Lusaka',
    'Africa/Malabo',
    'Africa/Maputo',
    'Africa/Maseru',
    'Africa/Mbabane',
    'Africa/Mogadishu',
    'Africa/Monrovia',
    'Africa/Nairobi',
    'Africa/Ndjamena',
    'Africa/Niamey',
    'Africa/Nouakchott',
    'Africa/Ouagadougou',
    'Africa/Porto-Novo',
    'Africa/Sao_Tome',
    'Africa/Tripoli',
    'Africa/Tunis',
    'Africa/Windhoek',
    'America/Adak',
    'America/Anchorage',
    'America/Anguilla',
    'America/Antigua',
    'America/Araguaina',
    'America/Argentina/Buenos_Aires',
    'America/Argentina/Catamarca',
    'America/Argentina/Cordoba',
    'America/Argentina/Jujuy',
    'America/Argentina/La_Rioja',
    'America/Argentina/Mendoza',
    'America/Argentina/Rio_Gallegos',
    'America/Argentina/Salta',
    'America/Argentina/San_Juan',
    'America/Argentina/San_Luis',
    'America/Argentina/Tucuman',
    'America/Argentina/Ushuaia',
    'America/Aruba',
    'America/Asuncion',
    'America/Atikokan',
    'America/Bahia',
    'America/Bahia_Banderas',
    'America/Barbados',
    'America/Belem',
    'America/Belize',
    'America/Blanc-Sablon',
    'America/Boa_Vista',
    'America/Bogota',
    'America/Boise',
    'America/Cambridge_Bay',
    'America/Campo_Grande',
    'America/Cancun',
    'America/Caracas',
    'America/Cayenne',
    'America/Cayman',
    'America/Chicago',
    'America/Chihuahua',
    'America/Ciudad_Juarez',
    'America/Costa_Rica',
    'America/Creston',
    'America/Cuiaba',
    'America/Curacao',
    'America/Danmarkshavn',
    'America/Dawson',
    'America/Dawson_Creek',
    'America/Denver',
    'America/Detroit',
    'America/Dominica',
    'America/Edmonton',
    'America/Eirunepe',
    'America/El_Salvador',
    'America/Fort_Nelson',
    'America/Fortaleza',
    'America/Glace_Bay',
    'America/Goose_Bay',
    'America/Grand_Turk',
    'America/Grenada',
    'America/Guadeloupe',
    'America/Guatemala',
    'America/Guayaquil',
    'America/Guyana',
    'America/Halifax',
    'America/Havana',
    'America/Hermosillo',
    'America/Indiana/Indianapolis',
    'America/Indiana/Knox',
    'America/Indiana/Marengo',
    'America/Indiana/Petersburg',
    'America/Indiana/Tell_City',
    'America/Indiana/Vevay',
    'America/Indiana/Vincennes',
    'America/Indiana/Winamac',
    'America/Inuvik',
    'America/Iqaluit',
    'America/Jamaica',
    'America/Juneau',
    'America/Kentucky/Louisville',
    'America/Kentucky/Monticello',
    'America/Kralendijk',
    'America/La_Paz',
    'America/Lima',
    'America/Los_Angeles',
    'America/Lower_Princes',
    'America/Maceio',
    'America/Managua',
    'America/Manaus',
    'America/Marigot',
    'America/Martinique',
    'America/Matamoros',
    'America/Mazatlan',
    'America/Menominee',
    'America/Merida',
    'America/Metlakatla',
    'America/Mexico_City',
    'America/Miquelon',
    'America/Moncton',
    'America/Monterrey',
    'America/Montevideo',
    'America/Montserrat',
    'America/Nassau',
    'America/New_York',
    'America/Nome',
    'America/Noronha',
    'America/North_Dakota/Beulah',
    'America/North_Dakota/Center',
    'America/North_Dakota/New_Salem',
    'America/Nuuk',
    'America/Ojinaga',
    'America/Panama',
    'America/Paramaribo',
    'America/Phoenix',
    'America/Port-au-Prince',
    'America/Port_of_Spain',
    'America/Porto_Velho',
    'America/Puerto_Rico',
    'America/Punta_Arenas',
    'America/Rankin_Inlet',
    'America/Recife',
    'America/Regina',
    'America/Resolute',
    'America/Rio_Branco',
    'America/Santarem',
    'America/Santiago',
    'America/Santo_Domingo',
    'America/Sao_Paulo',
    'America/Scoresbysund',
    'America/Sitka',
    'America/St_Barthelemy',
    'America/St_Johns',
    'America/St_Kitts',
    'America/St_Lucia',
    'America/St_Thomas',
    'America/St_Vincent',
    'America/Swift_Current',
    'America/Tegucigalpa',
    'America/Thule',
    'America/Tijuana',
    'America/Toronto',
    'America/Tortola',
    'America/Vancouver',
    'America/Whitehorse',
    'America/Winnipeg',
    'America/Yakutat',
    'Antarctica/Casey',
    'Antarctica/Davis',
    'Antarctica/DumontDUrville',
    'Antarctica/Macquarie',
    'Antarctica/Mawson',
    'Antarctica/McMurdo',
    'Antarctica/Palmer',
    'Antarctica/Rothera',
    'Antarctica/Syowa',
    'Antarctica/Troll',
    'Antarctica/Vostok',
    'Arctic/Longyearbyen',
    'Asia/Aden',
    'Asia/Almaty',
    'Asia/Amman',
    'Asia/Anadyr',
    'Asia/Aqtau',
    'Asia/Aqtobe',
    'Asia/Ashgabat',
    'Asia/Atyrau',
    'Asia/Baghdad',
    'Asia/Bahrain',
    'Asia/Baku',
    'Asia/Bangkok',
    'Asia/Barnaul',
    'Asia/Beirut',
    'Asia/Bishkek',
    'Asia/Brunei',
    'Asia/Chita',
    'Asia/Choibalsan',
    'Asia/Colombo',
    'Asia/Damascus',
    'Asia/Dhaka',
    'Asia/Dili',
    'Asia/Dubai',
    'Asia/Dushanbe',
    'Asia/Famagusta',
    'Asia/Gaza',
    'Asia/Hebron',
    'Asia/Ho_Chi_Minh',
    'Asia/Hong_Kong',
    'Asia/Hovd',
    'Asia/Irkutsk',
    'Asia/Jakarta',
    'Asia/Jayapura',
    'Asia/Jerusalem',
    'Asia/Kabul',
    'Asia/Kamchatka',
    'Asia/Karachi',
    'Asia/Kathmandu',
    'Asia/Khandyga',
    'Asia/Kolkata',
    'Asia/Krasnoyarsk',
    'Asia/Kuala_Lumpur',
    'Asia/Kuching',
    'Asia/Kuwait',
    'Asia/Macau',
    'Asia/Magadan',
    'Asia/Makassar',
    'Asia/Manila',
    'Asia/Muscat',
    'Asia/Nicosia',
    'Asia/Novokuznetsk',
    'Asia/Novosibirsk',
    'Asia/Omsk',
    'Asia/Oral',
    'Asia/Phnom_Penh',
    'Asia/Pontianak',
    'Asia/Pyongyang',
    'Asia/Qatar',
    'Asia/Qostanay',
    'Asia/Qyzylorda',
    'Asia/Riyadh',
    'Asia/Sakhalin',
    'Asia/Samarkand',
    'Asia/Seoul',
    'Asia/Shanghai',
    'Asia/Singapore',
    'Asia/Srednekolymsk',
    'Asia/Taipei',
    'Asia/Tashkent',
    'Asia/Tbilisi',
    'Asia/Tehran',
    'Asia/Thimphu',
    'Asia/Tokyo',
    'Asia/Tomsk',
    'Asia/Ulaanbaatar',
    'Asia/Urumqi',
    'Asia/Ust-Nera',
    'Asia/Vientiane',
    'Asia/Vladivostok',
    'Asia/Yakutsk',
    'Asia/Yangon',
    'Asia/Yekaterinburg',
    'Asia/Yerevan',
    'Atlantic/Azores',
    'Atlantic/Bermuda',
    'Atlantic/Canary',
    'Atlantic/Cape_Verde',
    'Atlantic/Faroe',
    'Atlantic/Madeira',
    'Atlantic/Reykjavik',
    'Atlantic/South_Georgia',
    'Atlantic/St_Helena',
    'Atlantic/Stanley',
    'Australia/Adelaide',
    'Australia/Brisbane',
    'Australia/Broken_Hill',
    'Australia/Darwin',
    'Australia/Eucla',
    'Australia/Hobart',
    'Australia/Lindeman',
    'Australia/Lord_Howe',
    'Australia/Melbourne',
    'Australia/Perth',
    'Australia/Sydney',
    'Canada/Atlantic',
    'Canada/Central',
    'Canada/Eastern',
    'Canada/Mountain',
    'Canada/Newfoundland',
    'Canada/Pacific',
    'Europe/Amsterdam',
    'Europe/Andorra',
    'Europe/Astrakhan',
    'Europe/Athens',
    'Europe/Belgrade',
    'Europe/Berlin',
    'Europe/Bratislava',
    'Europe/Brussels',
    'Europe/Bucharest',
    'Europe/Budapest',
    'Europe/Busingen',
    'Europe/Chisinau',
    'Europe/Copenhagen',
    'Europe/Dublin',
    'Europe/Gibraltar',
    'Europe/Guernsey',
    'Europe/Helsinki',
    'Europe/Isle_of_Man',
    'Europe/Istanbul',
    'Europe/Jersey',
    'Europe/Kaliningrad',
    'Europe/Kirov',
    'Europe/Kyiv',
    'Europe/Lisbon',
    'Europe/Ljubljana',
    'Europe/London',
    'Europe/Luxembourg',
    'Europe/Madrid',
    'Europe/Malta',
    'Europe/Mariehamn',
    'Europe/Minsk',
    'Europe/Monaco',
    'Europe/Moscow',
    'Europe/Oslo',
    'Europe/Paris',
    'Europe/Podgorica',
    'Europe/Prague',
    'Europe/Riga',
    'Europe/Rome',
    'Europe/Samara',
    'Europe/San_Marino',
    'Europe/Sarajevo',
    'Europe/Saratov',
    'Europe/Simferopol',
    'Europe/Skopje',
    'Europe/Sofia',
    'Europe/Stockholm',
    'Europe/Tallinn',
    'Europe/Tirane',
    'Europe/Ulyanovsk',
    'Europe/Vaduz',
    'Europe/Vatican',
    'Europe/Vienna',
    'Europe/Vilnius',
    'Europe/Volgograd',
    'Europe/Warsaw',
    'Europe/Zagreb',
    'Europe/Zurich',
    'GMT',
    'Indian/Antananarivo',
    'Indian/Chagos',
    'Indian/Christmas',
    'Indian/Cocos',
    'Indian/Comoro',
    'Indian/Kerguelen',
    'Indian/Mahe',
    'Indian/Maldives',
    'Indian/Mauritius',
    'Indian/Mayotte',
    'Indian/Reunion',
    'Pacific/Apia',
    'Pacific/Auckland',
    'Pacific/Bougainville',
    'Pacific/Chatham',
    'Pacific/Chuuk',
    'Pacific/Easter',
    'Pacific/Efate',
    'Pacific/Fakaofo',
    'Pacific/Fiji',
    'Pacific/Funafuti',
    'Pacific/Galapagos',
    'Pacific/Gambier',
    'Pacific/Guadalcanal',
    'Pacific/Guam',
    'Pacific/Honolulu',
    'Pacific/Kanton',
    'Pacific/Kiritimati',
    'Pacific/Kosrae',
    'Pacific/Kwajalein',
    'Pacific/Majuro',
    'Pacific/Marquesas',
    'Pacific/Midway',
    'Pacific/Nauru',
    'Pacific/Niue',
    'Pacific/Norfolk',
    'Pacific/Noumea',
    'Pacific/Pago_Pago',
    'Pacific/Palau',
    'Pacific/Pitcairn',
    'Pacific/Pohnpei',
    'Pacific/Port_Moresby',
    'Pacific/Rarotonga',
    'Pacific/Saipan',
    'Pacific/Tahiti',
    'Pacific/Tarawa',
    'Pacific/Tongatapu',
    'Pacific/Wake',
    'Pacific/Wallis',
    'US/Alaska',
    'US/Arizona',
    'US/Central',
    'US/Eastern',
    'US/Hawaii',
    'US/Mountain',
    'US/Pacific',
    'UTC'
]]


class TimeZoneHelper:
    
    def __init__(self, tz):
        
        if isinstance(tz, str):
            tz = ZoneInfo(tz)
        
        self.tz = tz
    
    @property
    def name(self):
        
        # ZoneInfo instances have a `key` attribute, but this also works for
        # the datetime.timezone.utc singleton
        return str(self.tz)
    
    def now(self):
        """
        Return the current datetime in the local timezone.
        """
        
        return timezone.now().astimezone(self.tz)
    
    def today(self):
        """
        Return the current date in the local timezone.
        """
        
        return self.now().date()
    
    def __str__(self):
        
        return self.name
    
    def __repr__(self):
        
        return f'<{self.__class__.__name__}: {self!s}>'


def get_tz_helper(value):
    """
    Return an instance of TimeZoneHelper based on the given value.
    Valid values are:
     - A timezone string (accepted by ZoneInfo())
     - An instance of ZoneInfo
     - The datetime.timezone.utc singleton
    
    Return None if the value is None or the empty string.
    Raise ValidationError if the value cannot be converted into a ZoneInfo instance.
    """
    
    if value in {None, ''}:
        return None
    
    if isinstance(value, TimeZoneHelper):
        return value
    
    try:
        return TimeZoneHelper(value)
    except ZoneInfoNotFoundError as e:
        msg = f'Invalid timezone "{value}".'
        raise ValidationError(msg) from e




class ModelAmbiguousVersionError(Exception):
    """
    Raised when a model's ``version`` field contains an ambiguous value, such
    as after it has been atomically incremented.
    """
    
    def __init__(self):
        
        super().__init__(
            'This instance\'s version has been atomically incremented and '
            'therefore its value, as associated with the current values of '
            'other fields on the instance, cannot be accurately determined.'
        )
    
    @classmethod
    def _raise(cls):
        
        raise cls()



import warnings

from django import forms
from django.contrib.auth import get_user_model
from django.test import TestCase

from djem.forms import (
    AuditableForm,
    CommonInfoForm,
    TimeZoneField,
    UserSavable,
)
from djem.utils.dt import TIMEZONE_CHOICES, TimeZoneHelper

from .models import StaticTest, TimeZoneTest


class UserSavableTestForm(UserSavable, forms.ModelForm):
    
    def __init__(self, *args, user=None, **kwargs):
        
        self.user = user
        
        super().__init__(*args, **kwargs)
    
    class Meta:
        model = StaticTest
        fields = '__all__'  # noqa: DJ007


class AuditableFormTestForm(AuditableForm):
    
    class Meta:
        model = StaticTest
        fields = '__all__'


class TimeZoneFieldTestForm1(forms.ModelForm):
    
    class Meta:
        model = TimeZoneTest
        fields = ['timezone']


class TimeZoneFieldTestForm2(forms.Form):
    
    # Manually defined field
    timezone = TimeZoneField()


class TimeZoneFieldTestForm3(forms.Form):
    
    # Manually defined field, not required
    timezone = TimeZoneField(required=False)


class TimeZoneFieldTestForm4(forms.Form):
    
    # Manually defined field with custom choices
    timezone = TimeZoneField(choices=TIMEZONE_CHOICES[:10])


class UserSavableTestCase(TestCase):
    
    @classmethod
    def setUpTestData(cls):
        
        cls.user = get_user_model().objects.create_user('test')
    
    def test_save_commit__true(self):
        """
        Test the form saves correctly when ``save`` is called with ``commit=True``
        (or no ``commit`` argument at all).
        """
        
        form = UserSavableTestForm({'test': True}, user=self.user)
        
        with self.assertNumQueries(1):
            instance = form.save()
        
        # Test correct insertion into database, with appropriate user fields
        self.assertIsNotNone(instance.pk)
        self.assertEqual(instance.user_created_id, self.user.pk)
        self.assertEqual(instance.user_modified_id, self.user.pk)
        
        # Test save_m2m method is inaccessible, as it should be after a
        # commit=True save
        with self.assertRaises(AttributeError):
            form.save_m2m  # noqa: B018
    
    def test_save_commit__false(self):
        """
        Test the form behaves correctly but does NOT save when ``save`` is
        called with ``commit=False``.
        """
        
        form = UserSavableTestForm({'test': True}, user=self.user)
        
        with self.assertNumQueries(0):
            instance = form.save(commit=False)
        
        # Test instance is not saved
        self.assertIsNone(instance.pk)
        
        # Test manual saving works
        instance.save(self.user)
        form.save_m2m()
        
        self.assertIsNotNone(instance.pk)
        self.assertEqual(instance.user_created_id, self.user.pk)
        self.assertEqual(instance.user_modified_id, self.user.pk)


class AuditableFormTestCase(TestCase):
    
    @classmethod
    def setUpTestData(cls):
        
        cls.user = get_user_model().objects.create_user('test')
    
    def test_bound_init__user(self):
        """
        Test the form correctly stores the given user when instantiating it
        with data (a bound form).
        """
        
        form = AuditableFormTestForm({'test': True}, user=self.user)
        
        self.assertEqual(form.user, self.user)
    
    def test_bound_init__no_user(self):
        """
        Test the form correctly raises a TypeError if no user is given when
        instantiating it with data (a bound form).
        """
        
        with self.assertRaises(TypeError):
            AuditableFormTestForm({'test': True})
    
    def test_unbound_init__user(self):
        """
        Test the form correctly stores the given user when instantiating it
        without data (an unbound form), even though it is not required.
        """
        
        form = AuditableFormTestForm(user=self.user)
        
        self.assertEqual(form.user, self.user)
    
    def test_unbound_init__no_user(self):
        """
        Test the form correctly stores ``None`` if no user is given when
        instantiating it without data (an unbound form).
        """
        
        form = AuditableFormTestForm()
        
        self.assertIsNone(form.user)


class CommonInfoFormTestCase(TestCase):
    
    def test_deprecation_warning(self):
        
        class TestForm(CommonInfoForm):
            
            class Meta:
                model = StaticTest
                fields = '__all__'
        
        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered
            warnings.simplefilter("always")
            
            TestForm()
            
            self.assertEqual(len(w), 1)
            self.assertIs(w[-1].category, DeprecationWarning)
            self.assertIn(
                'Use of CommonInfoForm is deprecated, use AuditableForm instead',
                str(w[-1].message)
            )


class TimeZoneFieldFormTestCase(TestCase):
    
    def test_submit__modelform_valid(self):
        """
        Test a ModelForm based on a Model with a TimeZoneField correctly
        accepts a valid submitted timezone string and cleans it to the correct
        TimeZoneHelper.
        """
        
        form = TimeZoneFieldTestForm1({'timezone': 'Australia/Sydney'})
        
        self.assertTrue(form.is_valid())
        self.assertIsInstance(form.cleaned_data['timezone'], TimeZoneHelper)
        self.assertEqual(form.cleaned_data['timezone'].name, 'Australia/Sydney')
    
    def test_submit__modelform_invalid(self):
        """
        Test a ModelForm based on a Model with a TimeZoneField correctly
        accepts an invalid submitted timezone string and generates the
        appropriate error.
        """
        
        form = TimeZoneFieldTestForm1({'timezone': 'fail'})
        
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['timezone'], [
            'Select a valid choice. fail is not one of the available choices.'
        ])
    
    def test_submit__form__valid(self):
        """
        Test a Form with a TimeZoneField correctly accepts a valid submitted
        timezone string and cleans it to the correct TimeZoneHelper.
        """
        
        form = TimeZoneFieldTestForm2({'timezone': 'Australia/Sydney'})
        
        self.assertTrue(form.is_valid())
        self.assertIsInstance(form.cleaned_data['timezone'], TimeZoneHelper)
        self.assertEqual(form.cleaned_data['timezone'].name, 'Australia/Sydney')
    
    def test_submit__form__invalid(self):
        """
        Test a Form with a TimeZoneField correctly accepts an invalid submitted
        timezone string and generates the appropriate error.
        """
        
        form = TimeZoneFieldTestForm2({'timezone': 'fail'})
        
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['timezone'], [
            'Select a valid choice. fail is not one of the available choices.'
        ])
    
    def test_submit__form__empty_valid__no_input(self):
        """
        Test a Form with a non-required TimeZoneField correctly accepts no input
        for the timezone value and cleans it to an empty string.
        """
        
        form = TimeZoneFieldTestForm3({})
        
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['timezone'], '')
    
    def test_submit__form__empty_valid__blank_string(self):
        """
        Test a Form with a non-required TimeZoneField correctly accepts an empty
        string as input for the timezone value and cleans it to an empty string.
        """
        
        form = TimeZoneFieldTestForm3({'timezone': ''})
        
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['timezone'], '')
    
    def test_submit__form__empty_valid__none(self):
        """
        Test a Form with a non-required TimeZoneField correctly accepts None as
        input for the timezone value and cleans it to an empty string.
        """
        
        form = TimeZoneFieldTestForm3({'timezone': None})
        
        self.assertTrue(form.is_valid())
        self.assertEqual(form.cleaned_data['timezone'], '')
    
    def test_submit__form__empty_invalid__no_input(self):
        """
        Test a Form with a required TimeZoneField does not accept no input for
        the timezone value and generates the appropriate error.
        """
        
        form = TimeZoneFieldTestForm2({})
        
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['timezone'], [
            'This field is required.'
        ])
    
    def test_submit__form__empty_invalid__blank_string(self):
        """
        Test a Form with a required TimeZoneField does not accept an empty
        string as input for the timezone value and generates the appropriate error.
        """
        
        form = TimeZoneFieldTestForm2({'timezone': ''})
        
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['timezone'], [
            'This field is required.'
        ])
    
    def test_submit__form__empty_invalid__none(self):
        """
        Test a Form with a required TimeZoneField does not accept None as input
        for the timezone value and generates the appropriate error.
        """
        
        form = TimeZoneFieldTestForm2({'timezone': None})
        
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['timezone'], [
            'This field is required.'
        ])
    
    def test_custom_choices__valid(self):
        """
        Test a Form with a TimeZoneField defined with a custom list of choices
        correctly accepts a valid submitted timezone string and cleans it to
        the correct TimeZoneHelper.
        """
        
        tz = TIMEZONE_CHOICES[0][0]
        
        form = TimeZoneFieldTestForm4({'timezone': tz})
        
        self.assertTrue(form.is_valid())
        self.assertIsInstance(form.cleaned_data['timezone'], TimeZoneHelper)
        self.assertEqual(form.cleaned_data['timezone'].name, tz)
    
    def test_custom_choices__invalid(self):
        """
        Test a Form with a TimeZoneField defined with a custom list of choices
        correctly accepts an invalid submitted timezone string and generates
        the appropriate error.
        """
        
        form = TimeZoneFieldTestForm4({'timezone': 'Australia/Sydney'})
        
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors['timezone'], [
            'Select a valid choice. Australia/Sydney is not one of the available choices.'
        ])



import warnings

from django import forms


class UserSavable:
    """
    A mixin for a Django ``ModelForm`` that adds support for models using the
    :class:`~djem.models.Auditable` model mixin. It handles providing a user
    instance to the model's ``save()`` method when the form's own ``save()``
    method is called, as is required by :class:`~djem.models.Auditable`.
    
    This mixin *assumes* the presence of a ``self.user`` attribute that the
    ``save()`` method can use. It is designed for use by forms that already
    accept and store a known user, e.g. as a constructor argument.
    
    For a form that provides the same customisation of the ``save()`` method
    and includes ``user`` as a constructor argument, see :class:`AuditableForm`.
    """
    
    def save(self, commit=True):
        
        # Call super.save() with commit=False so .save() can be called on the
        # instance with the required user argument
        instance = super().save(commit=False)
        
        if commit:
            instance.save(self.user)
            self.save_m2m()
            del self.save_m2m  # pretend commit=False was never used
        
        return instance


class AuditableForm(UserSavable, forms.ModelForm):
    """
    A Django ``ModelForm`` that is customised to support models using the
    :class:`~djem.models.Auditable` model mixin. It handles providing a user
    instance to the model's ``save()`` method when the form's own ``save()``
    method is called, as is required by :class:`~djem.models.Auditable`. It
    also adds a ``user`` keyword argument to the constructor so the ``save()``
    method has a known user to work with. The ``user`` argument is required if
    the field is bound, otherwise it is optional. The given user is stored in
    the :attr:`~AuditableForm.user` instance attribute. Subclasses may choose
    to use the known user for their own purposes.
    
    For a mixin that provides the same customisation of the ``save()`` method
    without the extra constructor argument (e.g. for use by forms that already
    accept and store a known user), see :class:`UserSavable`.
    """
    
    def __init__(self, *args, user=None, **kwargs):
        
        super().__init__(*args, **kwargs)
        
        # User is required for bound forms
        if self.is_bound and not user:
            msg = f'Bound {self.__class__.__name__} instances require a "user" argument.'
            raise TypeError(msg)
        
        # Set self.user regardless of whether or not the form is bound. Child
        # forms may well require the user when the form is unbound as well.
        self.user = user


# Backwards compat.
# TODO: Remove in 1.0
class CommonInfoForm(AuditableForm):
    
    def __init__(self, *args, **kwargs):
        
        warnings.warn(
            'Use of CommonInfoForm is deprecated, use AuditableForm instead.',
            DeprecationWarning,
            stacklevel=2
        )
        
        super().__init__(*args, **kwargs)



import re
import warnings
from collections import OrderedDict

from django.conf import settings
from django.contrib.auth.models import _user_has_perm  # noqa: PLC2701
from django.core.exceptions import ObjectDoesNotExist
from django.db import models, router, transaction
from django.db.models.deletion import Collector
from django.db.models.utils import resolve_callables
from django.utils import timezone
from django.utils.functional import SimpleLazyObject

from djem.exceptions import ModelAmbiguousVersionError

whitespace_regex = re.compile(r'\W+')

__all__ = (
    'Loggable', 'OLPMixin', 'MixableQuerySet',
    'Auditable', 'AuditableQuerySet', 'CommonInfoMixin', 'CommonInfoQuerySet',
    'Archivable', 'ArchivableQuerySet', 'ArchivableMixin',
    'Versionable', 'VersionableQuerySet', 'VersioningMixin', 'VersioningQuerySet',
    'StaticAbstract', 'StaticAbstractQuerySet',
)


def _is_user_required():
    
    # TODO: Remove fallback setting in 1.0
    return getattr(
        settings,
        'DJEM_AUDITABLE_REQUIRE_USER_ON_SAVE',
        getattr(settings, 'DJEM_COMMON_INFO_REQUIRE_USER_ON_SAVE', True)  # backwards compat.
    )


def _process_log(log, tags, raw):
    
    tags = set(tags) if tags else set()
    
    # Run the list comprehension whether or not there are tags to filter by,
    # so that the returned list is always a copy of the original
    log = [line for line in log if not tags or tags.intersection(line.tags)]
    
    if raw:
        return log
    
    return '\n'.join(log)


class UnarchivedCollector(Collector):
    
    def related_objects(self, related_model, related_fields, objs):
        """
        Overridden to filter out archived objects from the queryset used to
        collect related objects.
        """
        
        queryset = super().related_objects(related_model, related_fields, objs)
        
        # Filter out archived objects
        if issubclass(related_model, Archivable):
            queryset = queryset.filter(is_archived=False)
        
        return queryset


# str preferred over collections.UserString as parent class because the
# functionality being added is minimal and the goal is to be essentially
# indistinguishable from a regular string for most intents and purposes.
class _TaggableStr(str):  # noqa: FURB189
    
    def __new__(cls, value, tags=None):
        
        obj = super().__new__(cls, value)
        
        # Ensure tags are stored as a tuple to be consistent with the
        # immutability of string objects
        obj.tags = tuple(tags) if tags else ()
        
        return obj
    
    def __repr__(self):
        
        output = super().__repr__()
        
        if self.tags:
            tags = ','.join(sorted(self.tags))
            
            output = f'{output}, tags={tags}'
        
        return output


class Loggable:
    """
    A mixin for creating, storing, and retrieving logs on an instance. Named
    logs are stored internally on the ``Loggable`` instance and persist for the
    lifetime of the object. A single log is "active" at any given time and can
    be freely appended to while it is.
    """
    
    def __init__(self, *args, **kwargs):
        
        self._active_logs = OrderedDict()
        self._finished_logs = OrderedDict()
        
        super().__init__(*args, **kwargs)
    
    def start_log(self, name):
        """
        Start a new log with the given ``name``. The new log becomes the
        current "active" log. Queue any previous active log so that it can be
        reactivated when the new log is either finished or discarded.
        
        :param name: The name of the log.
        """
        
        if name in self._active_logs:
            msg = f'A log named "{name}" is already active.'
            raise ValueError(msg)
        
        self._active_logs[name] = []
    
    def end_log(self):
        """
        End the currently active log and return a ``(name, log)`` tuple, where
        ``name`` is the name of the log that was ended and ``log`` is a list of
        the entries that have been added to the log. Reactivate the previous
        log, if any.
        
        The returned list will be a *copy* of the one used to store the log
        internally, allowing it to be safely manipulated without affecting the
        original log.
        
        A log must be ended in order to be retrieved.
        
        :return: A ``(name, log)`` tuple.
        """
        
        # Pop from active logs to move to finished logs
        try:
            name, log = self._active_logs.popitem()
        except KeyError as e:
            msg = 'No active log to finish.'
            raise KeyError(msg) from e
        
        # If a log with the same name has been finished previously, remove it
        # from the finished logs dict before adding this one, so that this one
        # is added to the "end" of the ordered dict.
        if name in self._finished_logs:
            self._finished_logs.pop(name)
        
        self._finished_logs[name] = log
        
        return name, log.copy()
    
    def discard_log(self):
        """
        Discard the currently active log. Reactivate the previous log, if any.
        """
        
        try:
            self._active_logs.popitem()
        except KeyError as e:
            msg = 'No active log to discard.'
            raise KeyError(msg) from e
    
    def log(self, *lines, tag=None):
        """
        Append to the currently active log. Each given argument will be added
        as a separate line to the log. If ``tag`` is specified, each added line
        will be tagged with the given value.
        
        :param lines: Individual lines to add to the log.
        :param tag: A tag to apply to each line.
        """
        
        # Pop to get the last item
        try:
            name, log = self._active_logs.popitem()
        except KeyError as e:
            msg = 'No active log to append to. Has one been started?'
            raise KeyError(msg) from e
        
        # Append to the log
        tags = (tag, ) if tag else None
        for line in lines:
            log.append(_TaggableStr(line, tags=tags))
        
        # Put back in the active logs dict
        self._active_logs[name] = log
    
    def get_log(self, name, tags=None, raw=False):
        """
        Return the named log, as a string. The log must have been ended (via
        ``end_log()``) in order to retrieve it.
        
        Return a raw list of lines in the log if ``raw=True``. In this case,
        the returned list will be a *copy* of the one used to store the log
        internally, allowing it to be safely manipulated without affecting the
        original log.
        
        Whether returning a string or a list, use ``tags`` to filter the
        included lines to just those with at least one of the given tags.
        
        :param name: The name of the log to retrieve.
        :param tags: An iterable of tags to filter the log by.
        :param raw: ``True`` to return the log as a list. Returned as a string by default.
        :return: The log, either as a string or a list.
        """
        
        try:
            log = self._finished_logs[name]
        except KeyError as e:
            msg = f'No log found for "{name}". Has it been finished?'
            raise KeyError(msg) from e
        
        return _process_log(log, tags, raw)
    
    def get_last_log(self, tags=None, raw=False):
        """
        Return the most recently finished log, as a string.
        
        Return a raw list of lines in the log if ``raw=True``. In this case,
        the returned list will be a *copy* of the one used to store the log
        internally, allowing it to be safely manipulated without affecting the
        original log.
        
        Whether returning a string or a list, use ``tags`` to filter the
        included lines to just those with at least one of the given tags.
        
        :param tags: An iterable of tags to filter the log by.
        :param raw: ``True`` to return the log as a list. Returned as a string by default.
        :return: The log, either as a string or a list.
        """
        
        # Pop to get the last item, but put it straight back
        try:
            name, log = self._finished_logs.popitem()
        except KeyError as e:
            msg = 'No finished logs to retrieve.'
            raise KeyError(msg) from e
        
        self._finished_logs[name] = log
        
        return _process_log(log, tags, raw)


class OLPMixin(Loggable):
    """
    A companion to Django's ``PermissionsMixin`` that enables additional
    advanced features of the object-level permission system. It is not
    necessary to use this mixin in order to use object-level permissions,
    it just provides additional functionality (such as logging permission
    checks, optionally allowing superusers to be restricted by object-level
    conditions, etc).
    """
    
    #
    # NOTE: This does not *extend* `PermissionsMixin` so that it can be
    # used along with Django's `AbstractUser`, which already includes
    # `PermissionsMixin`.
    #
    
    def __init__(self, *args, **kwargs):
        
        self._olp_cache = {}
        
        super().__init__(*args, **kwargs)
    
    def _check_perm(self, perm, obj):
        
        if not getattr(settings, 'DJEM_UNIVERSAL_OLP', False):
            # Default behaviour: active superusers implicitly have ALL permissions
            if self.is_active and self.is_superuser:
                return True, 'Active superuser: Implicit permission'
            
            return _user_has_perm(self, perm, obj), None
        
        # "Universal OLP" behaviour: active superusers implicitly have all
        # permissions at the model level, but are subject to object-level
        # checks
        if not obj and self.is_active and self.is_superuser:
            return True, 'Active superuser: Implicit permission (model-level)'
        
        return _user_has_perm(self, perm, obj), None
    
    def logged_has_perm(self, perm, obj=None, verbosity=1):
        
        log_key = f'auto-{perm}'
        if obj:
            log_key = f'{log_key}-{obj.pk}'
        
        self.start_log(log_key)
        
        if verbosity > 1:
            log_lines = [
                f'Permission: {perm}',
                f'User: {self.get_username()} ({self.pk})'
            ]
            
            if obj:
                log_lines.append(f'Object: {obj!s} ({obj.pk})')
            
            log_lines.append('')  # blank line
            
            self.log(*log_lines, tag='auto')
        
        has_perm, log_entry = self._check_perm(perm, obj)
        
        if log_entry:
            self.log(log_entry, tag='auto')
        
        result = 'Permission Granted' if has_perm else 'Permission Denied'
        self.log(f'\nRESULT: {result}', tag='auto')
        self.end_log()
        
        return has_perm
    
    def has_perm(self, perm, obj=None):
        
        verbosity = getattr(settings, 'DJEM_PERM_LOG_VERBOSITY', 0)
        
        if verbosity:
            return self.logged_has_perm(perm, obj, verbosity)
        
        has_perm, _ = self._check_perm(perm, obj)
        return has_perm
    
    def clear_perm_cache(self):
        """
        Clear the object-level and model-level permissions caches on the user
        instance.
        """
        
        # Clear the object-level permissions cache
        self._olp_cache = {}
        
        # Clear the model-level permissions cache as well, for good measure
        try:
            del self._user_perm_cache
            del self._group_perm_cache
            del self._perm_cache
        except AttributeError:
            pass


class MixableQuerySet:
    """
    A mixin for ``QuerySet`` classes that simply provides an enhanced
    :meth:`~MixableQuerySet.as_manager` method that can be used to combine
    the queryset class with any number of other queryset classes automatically.
    """
    
    @classmethod
    def as_manager(cls, *other_querysets):
        """
        Similar to the ``as_manager`` classmethod `available on regular Django
        queryset classes
        <https://docs.djangoproject.com/en/stable/topics/db/managers/#creating-a-manager-with-queryset-methods>`_,
        this returns an instance of ``Manager`` with a copy of the queryset's
        methods. However, it also accepts *other* queryset classes as arguments
        and includes *their* methods in the created ``Manager`` also. This
        allows easily creating combinations of this queryset class with other
        custom queryset classes, without needing to manually create an extra
        class to do the grouping.
        
        This is only useful where the querysets being combined do not contain
        conflicting methods. Method inheritance is supported (i.e. multiple
        querysets can contain the same method and they will be resolved in
        normal method resolution order), but depending on the logic of those
        methods, they may not be compatible. In such cases, an extra class
        resolving any incompatibilities is still required.
        """
        
        #
        # Mostly copied from Django's QuerySet.as_manager(), but extended to
        # support combining querysets
        #
        
        queryset = cls
        base_name = cls.__name__
        
        # Slice "queryset" off the end of the base name if found. This just
        # makes the name a little nicer (and shorter) in common cases.
        if base_name.lower().endswith('queryset'):
            base_name = base_name[:-8]
        
        if other_querysets:
            # Create a single QuerySet that combines all given queryset classes
            base_name = f'{base_name}AndFriends'
            queryset = type(f'{base_name}QuerySet', (cls, *other_querysets), {})
        
        manager = models.Manager.from_queryset(queryset, f'{base_name}Manager')()
        manager._built_with_as_manager = True
        
        return manager
    as_manager.queryset_only = True


class AuditableQuerySet(MixableQuerySet, models.QuerySet):
    """
    Provides custom functionality pertaining to the fields provided by
    :class:`Auditable`.
    """
    
    def create(self, _user=None, **kwargs):
        """
        Overridden to ensure a user is provided to the ``save()`` call on the
        model instance. The ``_user`` argument (named to reduce potential
        conflicts with model field names) is the user instance to pass through.
        It is required unless the :setting:`DJEM_AUDITABLE_REQUIRE_USER_ON_SAVE`
        setting is ``False``.
        """
        
        if _is_user_required() and not _user:
            msg = 'create() requires the first positional argument to be a user model instance.'
            raise TypeError(msg)
        
        #
        # The below is copied from QuerySet.create() (as at Django 3.2.6) and
        # changed to pass the `user` argument to save()
        #
        
        obj = self.model(**kwargs)
        
        self._for_write = True
        obj.save(_user, force_insert=True, using=self.db)
        
        return obj
    
    def _extract_model_params(self, defaults, **kwargs):
        
        # Used by get_or_create() to combine `defaults` and `kwargs` to form
        # the params used by create(). Performs validation on fields, which
        # the injected `_user` param fails. So pop `_user` from `defaults`
        # before calling super(), and add it back afterwards.
        
        user = defaults.pop('_user')
        params = super()._extract_model_params(defaults, **kwargs)
        
        params['_user'] = user
        
        return params
    
    def get_or_create(self, defaults=None, _user=None, **kwargs):
        """
        Overridden to ensure a user is provided to the ``save()`` call on the
        model instance, if a record needs to be created. The ``_user`` argument
        (named to reduce potential conflicts with model field names) is the
        user instance to pass through. It is required unless the
        :setting:`DJEM_AUDITABLE_REQUIRE_USER_ON_SAVE` setting is ``False``.
        """
        
        # Add `_user` to `defaults` because `kwargs` are used in the lookup,
        # but `defaults` eventually become keyword arguments passed to `create()`
        if defaults is None:
            defaults = {}
        
        defaults['_user'] = _user
        
        return super().get_or_create(defaults, **kwargs)
    
    def update(self, _user=None, **kwargs):
        """
        Overridden to ensure the ``user_modified`` and ``date_modified`` fields
        are always updated. The ``_user`` argument (named to reduce potential
        conflicts with model field names) is the user instance to update
        ``user_modified`` with. It is required unless the
        :setting:`DJEM_AUDITABLE_REQUIRE_USER_ON_SAVE` setting is ``False``.
        """
        
        if _is_user_required() and not _user:
            msg = 'update() requires the first positional argument to be a user model instance.'
            raise TypeError(msg)
        
        kwargs.setdefault('date_modified', timezone.now())
        
        if _user:
            kwargs['user_modified'] = _user
        
        return super().update(**kwargs)
    
    def update_or_create(self, defaults=None, _user=None, **kwargs):
        """
        Overridden to ensure a user is provided to the ``save()`` call on the
        model instance, whether the record id being created or updated. The
        ``_user`` argument (named to reduce potential conflicts with model
        field names) is the user instance to pass through. It is required
        unless the :setting:`DJEM_AUDITABLE_REQUIRE_USER_ON_SAVE` setting
        is ``False``.
        """
        
        if _is_user_required() and not _user:
            msg = 'create() requires the first positional argument to be a user model instance.'
            raise TypeError(msg)
        
        # Add `_user` to `kwargs` rather than `defaults` as `_user` is its own
        # keyword argument to get_or_create() (which is called by super())
        kwargs['_user'] = _user
        
        #
        # The below is copied from QuerySet.update_or_create() (as at Django
        # 3.2.6) and changed to pass the `user` argument to save()
        #
        
        defaults = defaults or {}
        self._for_write = True
        with transaction.atomic(using=self.db):
            # Lock the row so that a concurrent update is blocked until
            # update_or_create() has performed its save.
            obj, created = self.select_for_update().get_or_create(defaults, **kwargs)
            if created:
                return obj, created
            
            for k, v in resolve_callables(defaults):
                setattr(obj, k, v)
                
            obj.save(_user, using=self.db)
        
        return obj, False
    
    def owned_by(self, user):
        """
        Return a queryset of records "owned" by the given user, as per the
        ``user_created`` field. ``user`` can be a ``User`` instance or an id.
        """
        
        return self.filter(user_created=user)


# Backwards compat.
# TODO: Remove in 1.0
class CommonInfoQuerySet(AuditableQuerySet):
    
    def __init__(self, *args, **kwargs):
        
        warnings.warn(
            'Use of CommonInfoQuerySet is deprecated, use AuditableQuerySet instead.',
            DeprecationWarning,
            stacklevel=2
        )
        
        super().__init__(*args, **kwargs)


class Auditable(models.Model):
    """
    Model mixin that provides standard user and datetime fields (``user_created``,
    ``user_modified``, ``date_created`` and ``date_modified``) and overridden
    instance and queryset methods to enforce keeping those details up to date.
    
    WARNING: Models incorporating this mixin cannot be involved in any process
    process that automatically calls ``save()`` on the instance, or many
    queryset methods that create/update record (``create()``, ``update()``,
    etc), as it won't pass the required user argument. Any such processes will
    require modification to support the custom method signatures, or the
    enforcement of a known user will need to be disabled.
    """
    
    date_created = models.DateTimeField(editable=False, verbose_name='Date Created')
    date_modified = models.DateTimeField(editable=False, verbose_name='Date Last Modified')
    user_created = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        verbose_name='User Created',
        related_name='+',
        on_delete=models.PROTECT
    )
    user_modified = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        verbose_name='User Last Modified',
        related_name='+',
        on_delete=models.PROTECT
    )
    
    objects = AuditableQuerySet.as_manager()
    
    class Meta:
        abstract = True
    
    def save(self, user=None, *args, **kwargs):
        """
        Overridden to ensure the ``user_modified`` and ``date_modified`` fields
        are always updated. The ``user`` argument is required and must be passed
        a ``User`` instance, unless the :setting:`DJEM_AUDITABLE_REQUIRE_USER_ON_SAVE`
        setting is ``False``.
        """
        
        if _is_user_required() and not user:
            msg = "save() requires the 'user' argument."
            raise TypeError(msg)
        
        now = timezone.now()
        update_fields = []
        
        self.date_modified = now
        update_fields.append('date_modified')
        
        if user:
            self.user_modified = user
            update_fields.append('user_modified')
        
        if self._state.adding:
            if self.date_created is None:
                self.date_created = now
            
            if user:
                try:
                    # Query for an existing user_created, replace if not found
                    self.user_created  # noqa: B018
                except ObjectDoesNotExist:
                    self.user_created = user
        
        if 'update_fields' in kwargs:
            # If only saving a subset of fields, make sure the fields altered
            # above are included. Not applicable when creating a new record,
            # so *_created fields can be ignored.
            kwargs['update_fields'] = set(kwargs['update_fields']).union(update_fields)
        
        super().save(*args, **kwargs)
    
    def owned_by(self, user):
        """
        Return ``True`` if ``user_created`` matches the given user, otherwise
        return ``False``. The user can be given as an id or a ``User`` instance.
        """
        
        try:
            user_id = user.pk
        except AttributeError:
            # Assume an id was given
            user_id = user
        
        return user_id == self.user_created_id


# Backwards compat.
# TODO: Remove in 1.0
class CommonInfoMixin(Auditable):
    
    class Meta:
        abstract = True
    
    def __init__(self, *args, **kwargs):
        
        warnings.warn(
            'Use of CommonInfoMixin is deprecated, use Auditable instead.',
            DeprecationWarning,
            stacklevel=2
        )
        
        super().__init__(*args, **kwargs)


class ArchivableQuerySet(MixableQuerySet, models.QuerySet):
    """
    Provides custom functionality pertaining to the ``is_archived`` field
    provided by :class:`Archivable`.
    """
    
    def archived(self):
        """
        Filter the queryset to archived records (``is_archived=True``).
        """
        
        return self.filter(is_archived=True)
    
    def unarchived(self):
        """
        Filter the queryset to unarchived records (``is_archived=False``).
        """
        
        return self.filter(is_archived=False)


class Archivable(models.Model):
    """
    Model mixin that provides an ``is_archived`` Boolean field, multiple
    Managers to access querysets filtered on that flag and additional instance
    and queryset methods to set the flag.
    """
    
    is_archived = models.BooleanField(default=False, db_index=True)
    
    objects = ArchivableQuerySet.as_manager()
    
    class Meta:
        abstract = True
    
    def archive(self, *args, **kwargs):
        """
        Archive this record.
        
        Accepts all arguments of the ``save`` method, as it saves the instance
        after setting the ``is_archived`` flag. It saves using the
        ``update_fields`` keyword argument, containing the ``is_archived``
        field, whether it was provided to this method or not. If provided, it
        is extended, not replaced.
        """
        
        # Collect objects referencing this one. Will raise ProtectedError or
        # RestrictedError if any references are through protected/restricted
        # foreign keys.
        using = kwargs.get('using') or router.db_for_write(self.__class__, instance=self)
        
        try:
            UnarchivedCollector(using=using).collect([self])
        except (models.ProtectedError, models.RestrictedError) as e:
            # Replace use of "delete" in the exception message with "archive"
            msg, objs = e.args
            msg = msg.replace('delete', 'archive')
            
            # Reconstruct the exceptions with the new message and the original
            # `protected_objects`/`restricted_objects` value
            raise type(e)(msg, objs) from None
        
        if 'update_fields' in kwargs:
            kwargs['update_fields'] = set(kwargs['update_fields']).union(('is_archived',))
        else:
            kwargs['update_fields'] = ('is_archived',)
        
        self.is_archived = True
        self.save(*args, **kwargs)
    
    def unarchive(self, *args, **kwargs):
        """
        Unarchive this record.
        
        Accepts all arguments of the ``save`` method, as it saves the instance
        after setting the ``is_archived`` flag. It saves using the
        ``update_fields`` keyword argument, containing the ``is_archived``
        field, whether it was provided to this method or not. If provided, it
        is extended, not replaced.
        """
        
        if 'update_fields' in kwargs:
            kwargs['update_fields'] = set(kwargs['update_fields']).union(('is_archived',))
        else:
            kwargs['update_fields'] = ('is_archived',)
        
        self.is_archived = False
        self.save(*args, **kwargs)


# Backwards compat.
# TODO: Remove in 1.0
class ArchivableMixin(Archivable):
    
    class Meta:
        abstract = True
    
    def __init__(self, *args, **kwargs):
        
        warnings.warn(
            'Use of ArchivableMixin is deprecated, use Archivable instead.',
            DeprecationWarning,
            stacklevel=2
        )
        
        super().__init__(*args, **kwargs)


class VersionableQuerySet(MixableQuerySet, models.QuerySet):
    """
    Provides custom functionality pertaining to the ``version`` field
    provided by :class:`Versionable`.
    """
    
    def update(self, **kwargs):
        """
        Overridden to ensure the ``version`` field is always updated.
        """
        
        kwargs['version'] = models.F('version') + 1
        
        return super().update(**kwargs)


# Backwards compat.
# TODO: Remove in 1.0
class VersioningQuerySet(VersionableQuerySet):
    
    def __init__(self, *args, **kwargs):
        
        warnings.warn(
            'Use of VersioningQuerySet is deprecated, use VersionableQuerySet instead.',
            DeprecationWarning,
            stacklevel=2
        )
        
        super().__init__(*args, **kwargs)


class Versionable(models.Model):
    """
    Model mixin that provides a ``version`` field that is automatically
    incremented on every save and overridden instance and manager methods to
    enforce keeping it up to date.
    
    WARNING: The version is atomically incremented using an F() expression. After
    a save, the field will no longer be accessible as an integer, and will raise
    AmbiguousVersionError if accessed. The instance will need to be requeried to
    get the new version. It can, however, be saved multiple times and the version
    will be correctly incremented each time.
    """
    
    # Model-specific version of the generic ModelAmbiguousVersionError exception
    class AmbiguousVersionError(ModelAmbiguousVersionError):
        pass
    
    version = models.PositiveIntegerField(editable=False, default=1)
    
    objects = VersionableQuerySet.as_manager()
    
    class Meta:
        abstract = True
    
    def save(self, *args, **kwargs):
        """
        Overridden to ensure the ``version`` field is always updated.
        """
        
        incremented = False
        
        if not self._state.adding:
            # Increment the version of this record. Does not happen on initial
            # save, as it is set to 1 by default.
            self.version = models.F('version') + 1
            incremented = True
            
            if 'update_fields' in kwargs:
                # If only saving a subset of fields, make sure the version
                # field is included.
                update_fields = set(kwargs['update_fields'])
                update_fields.add('version')
                kwargs['update_fields'] = update_fields
        
        super().save(*args, **kwargs)
        
        if incremented:
            # If the version has been incremented, make it inaccessible. It
            # cannot be accurately determined without re-querying for it, and
            # even getting an accurate version number does not mean it is the
            # version that correlates with the values of all other fields on
            # this instance. The only way to get all correlated values is to
            # re-query for the entire object. This is too much overhead to
            # impose on every save, especially when accessing the version after
            # a save will be an edge case. It will be up to application logic to
            # detect and handle the circumstance of an ambiguous version.
            self.version = SimpleLazyObject(self.AmbiguousVersionError._raise)


# Backwards compat.
# TODO: Remove in 1.0
class VersioningMixin(Versionable):
    
    class Meta:
        abstract = True
    
    def __init__(self, *args, **kwargs):
        
        warnings.warn(
            'Use of VersioningMixin is deprecated, use Versionable instead.',
            DeprecationWarning,
            stacklevel=2
        )
        
        super().__init__(*args, **kwargs)


class StaticAbstractQuerySet(AuditableQuerySet, ArchivableQuerySet, VersionableQuerySet):
    """
    Combination of AuditableQuerySet, ArchivableQuerySet and VersionableQuerySet
    for use by managers of models that want the functionality of all three.
    """
    
    pass


class StaticAbstract(Auditable, Archivable, Versionable, models.Model):
    """
    Useful abstract base model combining the functionality of the Auditable,
    Archivable, and Versionable mixins.
    """
    
    objects = StaticAbstractQuerySet.as_manager()
    
    class Meta:
        abstract = True



from django.db import models

from djem.utils.dt import TIMEZONE_CHOICES, get_tz_helper

__all__ = ('TimeZoneField', )


# Based on django-timezone-field
# https://github.com/mfogel/django-timezone-field
class TimeZoneField(models.Field):
    """
    Stores a timezone and provides a helper object to access date/time-related
    functionality for that timezone.
    
    Valid inputs:
     - A timezone string (accepted by ZoneInfo())
     - An instance of ZoneInfo
     - The datetime.timezone.utc singleton
     - An instance of djem.utils.dt.TimeZoneHelper
     - None and the empty string (both representing a null value)
    
    When the value of the field is not null, accessing the value of the field
    will output an instance of djem.utils.dt.TimeZoneHelper,
    instantiated with the stored timezone.
    """
    
    description = 'A timezone'
    
    CHOICES = TIMEZONE_CHOICES
    MAX_LENGTH = 63
    
    def __init__(self, verbose_name=None, **kwargs):
        
        kwargs.setdefault('choices', self.CHOICES)
        kwargs.setdefault('max_length', self.MAX_LENGTH)
        
        super().__init__(verbose_name=verbose_name, **kwargs)
    
    def get_internal_type(self):
        
        return 'CharField'
    
    def deconstruct(self):
        
        name, path, args, kwargs = super().deconstruct()
        
        # Only include choices and max_length kwargs if not the default
        if kwargs['choices'] == self.CHOICES:
            del kwargs['choices']
        
        if kwargs['max_length'] == self.MAX_LENGTH:
            del kwargs['max_length']
        
        return name, path, args, kwargs
    
    def get_default(self):
        
        # Convert string default to TimeZoneHelper
        value = super().get_default()
        return get_tz_helper(value)
    
    def from_db_value(self, value, expression, connection):
        
        # Convert to TimeZoneHelper
        return get_tz_helper(value)
    
    def to_python(self, value):
        
        # Convert to TimeZoneHelper
        return get_tz_helper(value)
    
    def get_prep_value(self, value):
        
        # Convert to timezone string, ensuring it is a valid timezone
        helper = get_tz_helper(value)
        
        if helper is None:
            if self.null:
                return None
            
            return ''
        
        return helper.name
    
    def validate(self, value, model_instance):
        
        # Ensure the value is a valid TimeZoneHelper for a valid timezone
        value = get_tz_helper(value)
        
        # Only pass the helper's timezone name into the super call - it will
        # be checked for its presence in self.choices
        super().validate(value.name, model_instance)



import re
import warnings

from django.apps import AppConfig, apps
from django.template import engines
from django.test import RequestFactory

from djem.middleware import MemoryStorage

BETWEEN_TAG_WHITESPACE_RE = re.compile(r'>\s+<')
EXCESS_WHITESPACE_RE = re.compile(r'\s\s+')


def setup_test_app(package, label=None):
    """
    Setup a Django test app for the provided package to allow test-only models
    to be used.
    
    This function should be called from ``myapp.tests.__init__`` like so::
    
        setup_test_app(__package__)
    
    This will create an app with the label "myapp_tests". If a specific app
    label is required, it can be provided explicitly::
    
        setup_test_app(__package__, 'mytests')
    
    Using either of the above, models can be placed in ``myapp.tests.models``
    and be discovered and used just like regular models by the test suite. As
    long as ``myapp.tests`` is not imported by anything that forms part of the
    standard Django runtime environment, these models will not be picked up in
    that environment, and will be isolated to the test suite only.
    """
    
    #
    # This solution is adapted from Simon Charette's comment on Django ticket #7835:
    # https://code.djangoproject.com/ticket/7835#comment:46
    #
    
    if label is None:
        containing_app_config = apps.get_containing_app_config(package)
        label = containing_app_config.label
        
        # Only suffix the app label if it has not been already. This allows
        # duplicate entries to be detected and prevented. It may prevent the
        # use of an implicit label if the tests reside in an app that
        # legitimately ends with "_tests", but an explicit label can always be
        # used. Without this check, earlier entries are returned by
        # get_containing_app_config() and suffixed repeatedly.
        if not containing_app_config.label.endswith('_tests'):
            label = f'{containing_app_config.label}_tests'
    
    if label in apps.app_configs:
        # An app with this label already exists, skip adding it. This is
        # necessary (vs raising an exception) as there are certain conditions
        # that can cause this function to be run multiple times (e.g. errors
        # during Django's initialisation can cause this).
        warnings.warn(f'Attempted setup of duplicate test app "{label}".', RuntimeWarning, stacklevel=2)
    
    app_config = AppConfig.create(package)
    app_config.apps = apps
    app_config.label = label
    
    apps.app_configs[label] = app_config
    
    app_config.import_models()
    
    apps.clear_cache()


class MessagingRequestFactory(RequestFactory):
    """
    An extension of Django's ``RequestFactory`` helper for tests that enables
    the use of the messages framework within the generated request. It does not
    use the standard message storage backend (as per the ``MESSAGE_STORAGE``
    setting), but rather a memory-only backend that does not involve the use
    of sessions, cookies or any other means of persistent storage of the
    messages. Thus, messages need to be read in the same request they were
    added, or they will be lost.
    """
    
    def request(self, **request):
        
        request = super().request(**request)
        
        # Patch the request so it appears as though the
        # djem.middleware.MessageMiddleware is in use, and it is an AJAX request
        request._messages = MemoryStorage(request)
        
        return request


class TemplateRendererMixin:
    """
    A mixin for TestCase classes whose tests render templates from strings (as
    opposed to rendering them from files), using the Django template engine.
    """
    
    def render_template(self, template_string, context, request=None, flatten=True):
        """
        Render the given string as a template, with the given context and
        request (if provided).
        
        If ``request`` is NOT provided, and a ``self.user`` attribute is
        available on the ``TestCase``, a "user" variable will be automatically
        added to the context.
        
        The rendered output will be stripped of any leading or trailing
        whitespace, and can optionally have excess whitespace "flattened" by
        passing the ``flatten`` argument as True (the default). Flattening
        removes ALL whitespace from between HTML tags and compresses all other
        whitespace down to a single space.
        
        :param template_string: The string to render as a template.
        :param context: The context with which to render the template.
        :param request: The ``HttpRequest`` with which to render the template context.
        :param flatten: True to "flatten" the rendered output (default), False
            to return the output with all internal whitespace intact.
        :return: The rendered template output.
        """
        
        if not request:
            try:
                context['user'] = self.user
            except AttributeError:
                pass
        
        output = engines['django'].from_string(template_string).render(context, request)
        
        if flatten:
            # Remove whitespace between tags
            output = re.sub(BETWEEN_TAG_WHITESPACE_RE, '><', output)
            
            # Replace instances of multiple whitespace characters with a single space
            output = re.sub(EXCESS_WHITESPACE_RE, ' ', output)
        
        return output.strip()  # remove unnecessary whitespace



from django.conf import settings
from django.core.paginator import EmptyPage, PageNotAnInteger, Paginator


def get_page(number, object_list, per_page=None, **kwargs):
    """
    Return the specified page, as a Django Page instance, from a Paginator
    constructed from the given object list and other keyword arguments.
    Handle InvalidPage exceptions and return logical valid pages instead.
    The ``per_page`` argument defaults to DJEM_DEFAULT_PAGE_LENGTH, if set.
    Otherwise, it is a required argument.
    """
    
    if per_page is None:
        try:
            per_page = settings.DJEM_DEFAULT_PAGE_LENGTH
        except AttributeError as e:
            msg = 'The "per_page" argument is required unless DJEM_DEFAULT_PAGE_LENGTH is set.'
            raise TypeError(msg) from e
    
    paginator = Paginator(object_list, per_page, **kwargs)
    
    try:
        return paginator.page(number)
    except PageNotAnInteger:
        number = 1
    except EmptyPage:
        if number < 1:
            # Page number too low, return first page
            number = 1
        elif paginator.num_pages:
            # Page number too high, return last page
            number = paginator.num_pages
        else:
            # Paginator has no pages, still try for the first page. Will return
            # an empty page unless allow_empty_first_page is False.
            number = 1
    
    return paginator.page(number)
